#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Classes for gathering segmented data.
"""
import datetime
import glob
import os.path
from ConfigParser import ConfigParser, NoSectionError, NoOptionError
from fnmatch import fnmatch
import numpy as np
import sys

import mpop.utils
from mpop import CONFIG_PATH
from mpop.scene import Satellite, SatelliteInstrumentScene
from mpop.projector import get_area_def
from mpop.satellites import GenericFactory

DEFAULT_TIMELINESS = datetime.timedelta(minutes=5)
DEFAULT_GRANULARITY = datetime.timedelta(minutes=5)


LOG = mpop.utils.get_logger("gatherer")

if sys.version_info < (2, 5):
    import time
    def strptime(string, fmt=None):
        """This function is available in the datetime module only
        from Python >= 2.5.
        """

        return datetime.datetime(*time.strptime(string, fmt)[:6])
else:
    strptime = datetime.datetime.strptime

    
def globify(filename):
    """Replace datetime string variable with ?'s.
    """
    filename = filename.replace("%Y", "????")
    filename = filename.replace("%j", "???")
    filename = filename.replace("%m", "??")
    filename = filename.replace("%d", "??")
    filename = filename.replace("%H", "??")
    filename = filename.replace("%M", "??")
    filename = filename.replace("%S", "??")
    return filename

def beginning(filename):
    """Find the beginning of *filename* not having any wildcards (? or *).
    Returns a duplet containing the position of the first wildcard, and the
    same position if the datetime variables where expanded.
    """
    posqm = filename.find("?")
    posst = filename.find("*")

    if posqm == -1 and posst == -1:
        pos = len(filename)
    elif posqm == -1:
        pos = posst
    elif posst == -1:
        pos = posqm
    else:
        pos = min(posst, posqm)

    secondpos = pos
    if filename[:pos].find("%Y") > -1:
        secondpos += 2
    if filename[:pos].find("%j") > -1:
        secondpos += 1
    return (pos, secondpos)

class Granule(SatelliteInstrumentScene):
    """The granule object.
    """

    def __init__(self, filename=None, time_slot=None,
                 satellite=None, instrument=None):

        # Setting up a granule from metadata

        if filename is None:
            SatelliteInstrumentScene.__init__(self, time_slot=time_slot,
                                              satellite=(satellite.satname,
                                                         satellite.number,
                                                         satellite.variant),
                                              instrument=instrument)
            conf = ConfigParser()
            conf.read(os.path.join(CONFIG_PATH, self.fullname + ".cfg"))
            self.file_template = str(conf.get(instrument+"-granules",
                                              "filename", raw=True))
            self.directory = str(conf.get(instrument+"-granules",
                                          "dir", raw=True))
            
            self.file_name = time_slot.strftime(self.file_template)
            self.directory = time_slot.strftime(self.directory)
            self.file_type = conf.get(instrument + "-granules", "type")
            self.granularity = datetime.timedelta(
                seconds=int(conf.get(instrument + "-granules", "granularity")))
            self.span = float(conf.get(instrument + "-granules",
                                       "full_scan_period"))
            self.scan_width = int(conf.get(instrument + "-granules",
                                           "scan_width"))
            return

        # Setting up a granule from a filename
        
        filelist = glob.glob(os.path.join(CONFIG_PATH, "*.cfg"))

        the_directory, the_name = os.path.split(filename)

        self.satname = None
        
        for fil in filelist:
            conf = ConfigParser()
            conf.read(fil)
            try:
                instruments = eval(conf.get("satellite", "instruments"))
                for instrument in instruments:
                    directory = str(conf.get(instrument+"-granules", "dir"))
                    file_template = str(conf.get(instrument+"-granules",
                                                 "filename", raw=True))
                    file_glob = globify(file_template)

                    if(os.path.samefile(the_directory, directory) and
                       fnmatch(the_name, file_glob)):
                        try:
                            self.file_template = file_template
                            self.file_name = the_name

                            pos1, pos2 = beginning(self.file_template)
                            time_slot = strptime(self.file_name[:pos2],
                                                 self.file_template[:pos1])
        
                            SatelliteInstrumentScene.__init__(
                                self,
                                time_slot=time_slot,
                                satellite=(conf.get("satellite", "satname"),
                                           conf.get("satellite", "number"),
                                           conf.get("satellite", "variant")),
                                instrument=instrument)

                            self.file_type = conf.get(instrument +
                                                      "-granules", "type")
                            self.directory = the_directory
                            self.granularity = datetime.timedelta(
                                seconds=int(conf.get(instrument + "-granules",
                                                     "granularity")))
                            self.span = float(conf.get(instrument +
                                                       "-granules",
                                                       "full_scan_period"))
                            self.scan_width = int(conf.get(instrument +
                                                           "-granules",
                                                           "scan_width"))
                        except (NoSectionError, NoOptionError):
                            raise IOError("Inconsistency detected in " + fil)
                        break

                if self.satname is not None:
                    break
                    
            except (NoSectionError, NoOptionError):
                pass

            
        if not self.satname:
            raise ValueError("Can't find any matching satellite for "+filename)
        

    def __cmp__(self, obj):
        return (cmp(self.satname, obj.satname) or
                cmp(self.number, obj.number) or
                cmp(self.variant, obj.variant) or
                cmp(self.time_slot, obj.time_slot))

    def __str__(self):
        return "G:" + os.path.join(self.directory, self.file_name)

    def get_lonlat(self, row, col):
        """Get the longitude and latitude for the current scene at the given
        row and col.
        """
        
        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, self.fullname + ".cfg"))

        try:
            reader_name = conf.get(self.instrument_name + "-level1", 'format')
        except NoSectionError:
            reader_name = conf.get(self.instrument_name + "-level2", 'format')
        reader = "mpop.satin." + reader_name
        try:
            reader_module = __import__(reader,
                                       globals(), locals(),
                                       ['get_lonlat'])
        except ImportError:
            LOG.exception("Problem finding a reader...")
            raise ImportError("No "+reader+" reader found.")
        return reader_module.get_lonlat(self, row, col)

    @property
    def gross_area(self):
        """Build a gross area of the segment based on its corners.
        """
        from pyresample.geometry import SwathDefinition

        nlines = self.granularity.seconds / self.span
        top_left = self.get_lonlat(0, 0)
        top_right = self.get_lonlat(0, self.scan_width - 1)
        bottom_left = self.get_lonlat(nlines - 1, 0)
        bottom_right = self.get_lonlat(nlines - 1, self.scan_width - 1)
        lons = np.array([[top_left[0], top_right[0]],
                         [bottom_left[0], bottom_right[0]]])
        lats = np.array([[top_left[1], top_right[1]],
                         [bottom_left[1], bottom_right[1]]])
        return SwathDefinition(lons, lats)
        
class SegmentedSwath(Satellite):

    def __init__(self, area, (satname, number, variant)):
        Satellite.__init__(self, (satname, number, variant))
        self.area = get_area_def(area)
        self.granules = []
        self.planned_granules = []
        self.timeout = None
    

    def add(self, granule):
        """Add a granule to the swath
        """
        if not self.granules:
            self._compute_plan(granule.time_slot,
                               granule.granularity,
                               granule.instrument_name)
        try:
            self.planned_granules.remove(granule)
        except ValueError:
            LOG.warning(str(granule) + " not in " +
                        str([str(gran) for gran in self.planned_granules]))
        self.granules.append(granule)
        if self.planned_granules:
            self.timeout = self.planned_granules[-1].time_slot
        else:
            self.timeout = datetime.datetime.now()

    def _compute_plan(self, utc_time, granularity, instrument):
        """Compute the planned granules for the current area.
        """
        nts = utc_time
        
        new_granule = Granule(time_slot=nts, satellite=self,
                              instrument=instrument)
        if new_granule.gross_area.overlaps(self.area):
            self.planned_granules.append(new_granule)

        while True:
            nts = nts - granularity
            new_granule = Granule(time_slot=nts, satellite=self,
                                  instrument=instrument)
            if new_granule.gross_area.overlaps(self.area):
                self.planned_granules.append(new_granule)
            else:
                break

        nts = utc_time
        while True:
            nts = nts + granularity
            new_granule = Granule(time_slot=nts, satellite=self,
                                  instrument=instrument)
            if new_granule.gross_area.overlaps(self.area):
                self.planned_granules.append(new_granule)
            else:
                break

        self.planned_granules.sort()
        self.timeout = self.planned_granules[-1].time_slot

    def __repr__(self):
        granules = [str(granule) for granule in self.granules]
        return "swath " + str(granules) + " on area " + self.area.area_id

    def concatenate(self, channels=None):
        """Concatenate the granules into a swath. Returns a loaded satellite
        scene.
        """
        self.granules.sort()
        conffile = (self.granules[0].variant +
                    self.granules[0].satname +
                    self.granules[0].number +
                    ".cfg")
        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, conffile))

        try:
            
            # call concat

            reader_name = conf.get(self.granules[0].instrument_name +
                                   "-level1", 'format')

            str_granules = [str(granule) for granule in self.granules]
            LOG.debug("Concatenating "+str(str_granules))

            try:
                reader_name = eval(reader_name)
            except NameError:
                reader_name = str(reader_name)

            # read the data
            reader = "mpop.satin."+reader_name

            try:
                reader_module = __import__(reader, globals(), locals(),
                                           ['concatenate'])
            except ImportError:
                LOG.exception("ImportError when loading plugin for format " +
                              str(reader_name))
                raise ImportError("No "+reader+" reader found.")

            scene = reader_module.concatenate(self.granules, channels)
            
        except NoSectionError:
            #concatenate loaded granules.
            scenes = [GenericFactory.create_scene(granule.satname,
                                                  granule.number,
                                                  granule.instrument_name,
                                                  granule.time_slot,
                                                  None,
                                                  None,
                                                  granule.variant)
                      for granule in self.granules]
            for granule in scenes:
                granule.load(channels)
            scene = mpop.scene.assemble_segments(scenes)

        return scene
        
            
class Gatherer(Satellite):
    """The mighty gatherer class.
    """

    def __init__(self, areas_of_interest,
                 timeliness=DEFAULT_TIMELINESS,
                 satname=None, number=None, variant=None):
        Satellite.__init__(self, (satname, number, variant))
        self.timeliness = timeliness
        self.swaths = {}
        self.finished_swaths = []


        for area in areas_of_interest:
            self.swaths[area] = SegmentedSwath(area, (satname, number, variant))

    def add(self, granule):
        """Add a *granule* to the gatherer.
        """
        interesting = False
        for area_name in self.swaths:
            if granule.gross_area.overlaps(self.swaths[area_name].area):
                interesting = True
                if (len(self.swaths[area_name].planned_granules) > 0 and
                    granule not in self.swaths[area_name].planned_granules):
                    LOG.debug("Following swath is starting, closing the current"
                              " swath")
                    LOG.debug("Swath was " + str(self.swaths[area_name]) +
                              " while " +
                              str(self.swaths[area_name].planned_granules) +
                              " remains.")
                    self.finished_swaths.append(self.swaths[area_name])
                    self.swaths[area_name] = SegmentedSwath(area_name,
                                                            (granule.satname,
                                                             granule.number,
                                                             granule.variant))
                LOG.debug("Add " + str(granule) + " to " +
                          str(self.swaths[area_name]))
                self.swaths[area_name].add(granule)
                if len(self.swaths[area_name].planned_granules) == 0:
                    self.finished_swaths.append(self.swaths[area_name])
                    self.swaths[area_name] = SegmentedSwath(area_name,
                                                            (granule.satname,
                                                             granule.number,
                                                             granule.variant))
        return interesting
    
    def timeout(self):
        """Finishes swaths that are timed out.
        """
        now = datetime.datetime.utcnow()
        
        for area_name, swath in self.swaths.items():
            if swath.timeout and swath.timeout + self.timeliness < now:
                self.finished_swaths.append(swath)
                self.swaths[area_name] = SegmentedSwath(area_name,
                                                        (self.satname,
                                                         self.number,
                                                         self.variant))
    def clear(self):
        """Clean up the finished swaths.
        """
        self.finished_swaths = []
        
    def __str__(self):
        return self.fullname
        
# Typical example:

# when new_file:
#     gatherer.add(Granule(new_file))
#     for swath in gatherer.finished:
#         scene = swath.assemble()
#         do_something_with(scene)
#     gatherer.finished.clear()

    
# when its time (gatherer.next_timeout):
#     gatherer.finish_timeouts()
#     for swath in gatherer.finished:
#         scene = swath.assemble()
#         do_something_with(scene)
#     gatherer.finished.clear()


import unittest

OldConfigParser = None

def patch_glob():
    """Patch fnmatch.
    """
    def fake_glob(*args):
        """Fake glob.
        """
        del args
        return ["test.cfg"]

    glob.old_glob = glob.glob
    glob.glob = fake_glob

def unpatch_glob():
    """Unpatch glob.
    """
    glob.glob = glob.old_glob
    delattr(glob, "old_glob")

def patch_configparser():
    """Patch to fake ConfigParser.
    """
    class FakeConfigParser:
        """Dummy ConfigParser class.
        """
        def __init__(self, *args, **kwargs):
            pass
        
        def read(self, *args, **kwargs):
            """Dummy read method
            """
            del args, kwargs
            self = self

        def get(self, section, item, **kwargs):
            """Dummy get method
            """
            del kwargs
            self = self
            if(section == "avhrr-granules" and
               item == "filename"):
                return "myfile_%Y%m%d_%H%M.cooltype"
            if(section == "avhrr-granules" and
               item == "type"):
                return "cooltype"
            if(section == "avhrr-granules" and
               item == "granularity"):
                return 60
            if(section == "avhrr-granules" and
               item == "full_scan_period"):
                return 0.1667
            if(section == "avhrr-granules" and
               item == "scan_width"):
                return 2048
            if(section == "avhrr-granules" and
               item == "dir"):
                return "/path/to/my/data"
            if(section == "satellite" and
               item == "instruments"):
                return "('avhrr', )"
            if(section == "satellite" and
               item == "satname"):
                return "metop"
            if(section == "satellite" and
               item == "number"):
                return "02"
            if(section == "satellite" and
               item == "variant"):
                return "regional"
            

    global OldConfigParser, ConfigParser

    OldConfigParser = ConfigParser
    ConfigParser = FakeConfigParser

def unpatch_configparser():
    """Unpatch fake ConfigParser.
    """
    global ConfigParser
    ConfigParser = OldConfigParser
    



class TestGranules(unittest.TestCase):
    """Testing granules.
    """

    def setUp(self):
        """Patching stuff.
        """
        patch_configparser()
        patch_glob()

    def tearDown(self):
        """Unpatching stuff.
        """
        unpatch_configparser()
        unpatch_glob()

    def test_init(self):
        """Testing init function.
        """

        g_1 = Granule("/path/to/my/data/myfile_20101010_1010.cooltype")

        self.assertEquals(g_1.satname, "metop")
        self.assertEquals(g_1.number, "02")
        self.assertEquals(g_1.variant, "regional")
        self.assertEquals(g_1.time_slot,
                          datetime.datetime(2010, 10, 10, 10, 10))
        self.assertEquals(g_1.instrument_name, "avhrr")
        
        sat = Satellite("metop", "02", "regional")
        g_1 = Granule(instrument="avhrr", satellite=sat,
                     time_slot=datetime.datetime(2010, 10, 10, 10, 10))
        
        self.assertEquals(os.path.join(g_1.directory, g_1.file_name),
                          "/path/to/my/data/myfile_20101010_1010.cooltype")
        

    def test_cmp(self):
        """test the __cmp__ function.
        """
        g_1 = Granule("/path/to/my/data/myfile_20101010_1010.cooltype")
        g_2 = Granule("/path/to/my/data/myfile_20101010_1011.cooltype")
        self.assertTrue(g_1 < g_2)
        self.assertTrue(g_2 > g_1)
        self.assertTrue(g_1 == g_1)


OldGranule = None

def patch_granule():
    """Faking Granule.
    """
    class FakeArea:
        """Fake area class.
        """
        def __init__(self, inside):
            self.inside = inside
            
        def overlaps(self, other):
            """Fake overlaping function.
            """
            del other
            return self.inside
    
    class FakeGranule:
        """Fake granule class.
        """
        def __init__(self, time_slot, satellite):
            self.time_slot = time_slot
            self.satellite = satellite
            self.span = 1

        def __cmp__(self, other):
            return cmp(self.time_slot, other.time_slot)

        @property
        def gross_area(self):
            """Approximate area of the granule.
            """
            if(self.time_slot > 2 and
               self.time_slot < 8):
                return FakeArea(inside=True)
            else:
                return FakeArea(inside=False)

        def __repr__(self):
            return "G:" + str(self.time_slot)
        
    global Granule, OldGranule
    OldGranule = Granule
    Granule = FakeGranule

def patch_granule_with_time():
    """Faking Granule.
    """
    class FakeArea:
        """Fake area class.
        """
        def __init__(self, inside):
            self.inside = inside
            
        def overlaps(self, other):
            """Fake overlaping function.
            """
            return self.inside
    
    class FakeGranule:
        """Fake granule class.
        """
        satname = "bli"
        number = "blu"
        variant = "bla"
        def __init__(self, time_slot, satellite):
            self.time_slot = time_slot
            self.satellite = satellite
            self.span = datetime.timedelta(minutes=1)

        def __cmp__(self, other):
            return cmp(self.time_slot, other.time_slot)

        @property
        def gross_area(self):
            """Approximate area of the granule.
            """
            start_time = datetime.datetime(2010, 10, 10, 0, 2)
            end_time = datetime.datetime(2010, 10, 10, 0, 8)
            if(self.time_slot > start_time and
               self.time_slot < end_time):
                return FakeArea(inside=True)
            else:
                return FakeArea(inside=False)

        def __repr__(self):
            return "G:" + str(self.time_slot)
        
    global Granule, OldGranule
    OldGranule = Granule
    Granule = FakeGranule

def unpatch_granule():

    global Granule
    Granule = OldGranule
    

class TestSegmentedSwath(unittest.TestCase):
    """Testing SegmentedSwath.
    """
    def setUp(self):
        patch_granule()

    def tearDown(self):
        unpatch_granule()

    def test_init(self):
        """Test initialisation.
        """
        swath = SegmentedSwath("bla", "bli", "blu", "blo")
        self.assertEquals(swath.area, "bla")

    def test_add(self):
        """Test adding.
        """
        swath = SegmentedSwath("bla", "bli", "blu", "blo")
        granule = Granule(5, "kurt")
        
        swath.add(granule)
        self.assertEquals(granule.time_slot, swath.granules[0].time_slot)
        times = [granule.time_slot for granule in swath.planned_granules]
        self.assertEquals(times, [3, 4, 6, 7])

        self.assertEquals(swath.timeout, 7)
        
        granule = Granule(4, "kurt")
        swath.add(granule)

        times = [granule.time_slot for granule in swath.planned_granules]
        self.assertEquals(times, [3, 6, 7])

        times = [granule.time_slot for granule in swath.granules]
        times.sort()
        self.assertEquals(times, [4, 5])

        self.assertEquals(swath.timeout, 7)
        
        granule = Granule(6, "kurt")
        swath.add(granule)
        
        times = [granule.time_slot for granule in swath.planned_granules]
        self.assertEquals(times, [3, 7])

        times = [granule.time_slot for granule in swath.granules]
        times.sort()
        self.assertEquals(times, [4, 5, 6])

        self.assertEquals(swath.timeout, 7)

        granule = Granule(7, "kurt")
        swath.add(granule)

        times = [granule.time_slot for granule in swath.planned_granules]
        self.assertEquals(times, [3])

        times = [granule.time_slot for granule in swath.granules]
        times.sort()
        self.assertEquals(times, [4, 5, 6, 7])

        self.assertEquals(swath.timeout, 3)

        granule = Granule(3, "kurt")
        swath.add(granule)

        times = [granule.time_slot for granule in swath.planned_granules]
        self.assertEquals(times, [])

        times = [granule.time_slot for granule in swath.granules]
        times.sort()
        self.assertEquals(times, [3, 4, 5, 6, 7])

        self.assertTrue((datetime.datetime.now() - swath.timeout).seconds < 1)
        self.assertTrue(isinstance(swath.timeout, datetime.datetime))



    def test_compute_plan(self):
        """Test planning of comming granules.
        """
        swath = SegmentedSwath("bla", "bli", "blu", "blo")

        swath._compute_plan(5, 1)
        times = [granule.time_slot for granule in swath.planned_granules]
        self.assertEquals(times, [3, 4, 5, 6, 7])


def patch_now():
    """Patching the now function from datetime.
    """
    def fakenow():
        """Fake now function.
        """
        return datetime.datetime(2010, 10, 10, 0, 10)
    
    datetime.datetime.oldnow = datetime.datetime.now
    datetime.datetime.now = fakenow

def unpatch_now():
    """Returning the original now.
    """
    datetime.datetime.now = datetime.datetime.oldnow
    delattr(datetime.datetime, "oldnow")

class TestGatherer(unittest.TestCase):
    """Testing Gatherer.
    """

    def setUp(self):
        patch_granule_with_time()
        
    def tearDown(self):
        unpatch_granule()
        
    def test_init(self):
        """Testing initialisation.
        """
        gatherer = Gatherer(["bli", "blu"],
                            satname="gla", number="glo", variant="glu")
        self.assertTrue("bli" in gatherer.swaths)
        self.assertTrue(isinstance(gatherer.swaths["bli"], SegmentedSwath))
        self.assertTrue("blu" in gatherer.swaths)
        self.assertTrue(isinstance(gatherer.swaths["blu"], SegmentedSwath))

    def test_add(self):
        """Testing adding of granules to the gatherer.
        """
        timeliness = (datetime.datetime.utcnow() -
                      datetime.datetime(2010, 10, 9, 23, 0))

        gatherer = Gatherer(["bli", "blu"],
                            timeliness=timeliness,
                            satname="gla", number="glo", variant="glu")
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 5), "blaf"))
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 3), "blaf"))
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 4), "blaf"))

        times = [granule.time_slot
                 for granule in gatherer.swaths["bli"].granules]
        ideal_times = [datetime.datetime(2010, 10, 10, 0, 5),
                       datetime.datetime(2010, 10, 10, 0, 3),
                       datetime.datetime(2010, 10, 10, 0, 4)]
        self.assertEquals(times, ideal_times)

        # put the timeliness to zero and add a new granule:
        # This should add the granule as normal.
        
        gatherer.timeliness = datetime.timedelta(seconds=0)
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 7), "blaf"))

        times = [granule.time_slot
                 for granule in gatherer.swaths["bli"].granules]
        ideal_times = [datetime.datetime(2010, 10, 10, 0, 5),
                       datetime.datetime(2010, 10, 10, 0, 3),
                       datetime.datetime(2010, 10, 10, 0, 4),
                       datetime.datetime(2010, 10, 10, 0, 7)]
        self.assertEquals(times, ideal_times)

        gatherer.timeliness = timeliness
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 6), "blaf"))

        self.assertFalse(gatherer.swaths["bli"].granules)
        self.assertFalse(gatherer.swaths["blu"].granules)
        self.assertEquals(len(gatherer.finished_swaths), 2)


    def test_timeout(self):
        """Test the timeout.
        """
        timeliness = (datetime.datetime.utcnow() -
                      datetime.datetime(2010, 10, 9, 23, 0))

        gatherer = Gatherer(["kurt", "blu"],
                            timeliness=timeliness,
                            satname="gla", number="glo", variant="glu")
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 5), "blaf"))
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 3), "blaf"))
        gatherer.add(Granule(datetime.datetime(2010, 10, 10, 0, 4), "blaf"))

        gatherer.timeout()
        times = [granule.time_slot
                 for granule in gatherer.swaths["kurt"].granules]
        ideal_times = [datetime.datetime(2010, 10, 10, 0, 5),
                       datetime.datetime(2010, 10, 10, 0, 3),
                       datetime.datetime(2010, 10, 10, 0, 4)]
        self.assertEquals(times, ideal_times)

        gatherer.timeliness = datetime.timedelta(seconds=0)
        gatherer.timeout()
        times = [granule.time_slot
                 for granule in gatherer.finished_swaths[0].granules]
        ideal_times = [datetime.datetime(2010, 10, 10, 0, 5),
                       datetime.datetime(2010, 10, 10, 0, 3),
                       datetime.datetime(2010, 10, 10, 0, 4)]
        self.assertEquals(times, ideal_times)

        self.assertFalse(gatherer.swaths["kurt"].granules)
        self.assertFalse(gatherer.swaths["kurt"].planned_granules)
        self.assertFalse(gatherer.swaths["blu"].granules)
        self.assertFalse(gatherer.swaths["blu"].planned_granules)
        

        
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions to read Grib messages
"""

import os
import pygrib
import os.path


class Grib(object):

    def __init__(self, fname):

        self._abspath = os.path.abspath(fname)

    @property
    def nmsgs(self):
        '''Number of GRIB messages in file.
        '''

        prop = 'nmsgs'
        attr = '_{}'.format(prop)

        if not hasattr(self, attr):
            grbs = pygrib.open(self._abspath)
            nmsgs = grbs.messages
            grbs.close()

            setattr(self, attr, nmsgs)

        return getattr(self, attr)

    def get(self, gmessage, key='values'):
        '''
        Returns the value for the 'key' for a given message number 'gmessage' or
        message field name 'gmessage'.
        '''

        grbs = pygrib.open(self._abspath)

        if type(gmessage) == int:
            mnbr = gmessage
        elif type(gmessage) == str:
            msg_found = False
            msgnum = 1
            while msgnum < self.nmsgs + 1:
                if grbs[msgnum]['parameterName'] == gmessage:
                    msg_found = True
                    break
                msgnum = msgnum + 1

            if msg_found:
                mnbr = msgnum
            else:
                print("No Grib message found with parameter name = %s" %
                      gmessage)
                return None

        if grbs[mnbr].valid_key(key):

            arr = grbs[mnbr][key]
            grbs.close()
            return arr
        else:
            grbs.close()
            return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2014, 2016.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Ronald Scheirer <ronald.scheirer@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to Modis level 1b format send through Eumetcast.
http://www.icare.univ-lille1.fr/wiki/index.php/MODIS_geolocation
http://www.sciencedirect.com/science?_ob=MiamiImageURL&_imagekey=B6V6V-4700BJP-\
3-27&_cdi=5824&_user=671124&_check=y&_orig=search&_coverDate=11%2F30%2F2002&vie\
w=c&wchp=dGLzVlz-zSkWz&md5=bac5bc7a4f08007722ae793954f1dd63&ie=/sdarticle.pdf
"""
import glob
from fnmatch import fnmatch
import os.path
from ConfigParser import ConfigParser
import multiprocessing
from trollsift.parser import Parser, globify

import math
import numpy as np
from pyhdf.SD import SD
from pyhdf.error import HDF4Error
import hashlib
from pyresample import geometry
import copy

from mpop import CONFIG_PATH
from mpop.plugin_base import Reader
from mpop.scene import assemble_segments

import logging
logger = logging.getLogger(__name__)


def get_filename(template, time_slot):
    tmpl = time_slot.strftime(template)
    file_list = glob.glob(tmpl)
    if len(file_list) > 1:
        raise IOError("More than 1 file matching template %s", tmpl)
    elif len(file_list) == 0:
        raise IOError("No EOS MODIS file matching " + tmpl)
    return file_list[0]


def check_filename(tmpl):
    file_list = glob.glob(tmpl)
    if len(file_list) > 1:
        raise IOError("More than 1 file matching template %s", tmpl)
    elif len(file_list) == 0:
        raise IOError("No EOS MODIS file matching " + tmpl)
    return file_list[0]


class ModisReader(Reader):

    pformat = "hdfeos_l1b"

    res = {"1": 1000,
           "Q": 250,
           "H": 500}

    inv_res = {1000: "1",
               250: "Q",
               500: "H"}

    def __init__(self, *args, **kwargs):
        Reader.__init__(self, *args, **kwargs)
        self.datafiles = {}
        self.geofile = None
        self.filename = None
        self.data = None
        self.areas = {}

    def get_sunsat_angles(self, resolution=1000):
        """Get sun-satellite viewing geometry for the 1km resolution data
        Optional arguments:
            None
        Return
            sun-zenith, sun-azimuth, sat-zenith, sat-azimuth

        """
        logger.debug("generating sun-sat viewing angles at %d", resolution)
        if self.geofile is not None:
            coarse_resolution = 1000
            filename = self.geofile
        else:
            coarse_resolution = 5000
            logger.info("Using 5km Sun-Sat viewing geometry and interpolating")
            filename = (self.datafiles.get(1000) or
                        self.datafiles.get(500) or
                        self.datafiles.get(250))
            raise NotImplementedError("Not yet implemented...")

        logger.debug("Loading sun-sat angles from file: " + str(filename)
                     + " at resolution " + str(coarse_resolution))

        eosdata = SD(str(filename))
        hdf_names = ['SolarZenith', 'SolarAzimuth',
                     'SensorZenith', 'SensorAzimuth']
        local_names = ['sunz', 'sun_azi',
                       'satz', 'sat_azi']
        data = {}
        for lname, dname in zip(local_names, hdf_names):
            data[lname] = eosdata.select(dname)
            fill_value = data[lname].attributes()["_FillValue"]
            scale = data[lname].attributes()["scale_factor"]
            data[lname] = np.ma.masked_equal(data[lname].get(), fill_value)
            data[lname] = data[lname] * scale

        return (data['sunz'], data['sun_azi'],
                data['satz'], data['sat_azi'])

    def load(self, satscene, filename=None, *args, **kwargs):
        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
        options = dict(conf.items(satscene.instrument_name + "-level2",
                                  raw=True))
        options["resolution"] = 1000
        options["geofile"] = os.path.join(options["dir"], options["geofile"])
        options.update(kwargs)

        fparser = Parser(options.get("filename"))
        gparser = Parser(options.get("geofile"))

        if filename is not None:
            datasets = {}
            if not isinstance(filename, (list, set, tuple)):
                filename = [filename]

            for fname in filename:
                if fnmatch(os.path.basename(fname), fparser.globify()):
                    metadata = fparser.parse(os.path.basename(fname))
                    datasets.setdefault(
                        metadata["start_time"], []).append(fname)
                elif fnmatch(os.path.basename(fname), gparser.globify()):
                    metadata = fparser.parse(fname)
                    datasets.setdefault(
                        metadata["start_time"], []).append(fname)

            scenes = []
            for start_time, dataset in datasets.iteritems():
                newscn = copy.deepcopy(satscene)
                newscn.time_slot = start_time
                self.load_dataset(newscn, filename=dataset, *args, **kwargs)
                scenes.append(newscn)

            if not scenes:
                logger.debug("Looking for files")
                self.load_dataset(satscene, *args, **kwargs)
            else:
                entire_scene = assemble_segments(
                    sorted(scenes, key=lambda x: x.time_slot))
                satscene.channels = entire_scene.channels
                satscene.area = entire_scene.area
                satscene.orbit = int(entire_scene.orbit)
                satscene.info["orbit_number"] = int(entire_scene.orbit)
        else:
            self.load_dataset(satscene, *args, **kwargs)

    def load_dataset(self, satscene, filename=None, *args, **kwargs):
        """Read data from file and load it into *satscene*.
        """
        del args
        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
        options = dict(conf.items(satscene.instrument_name + "-level2",
                                  raw=True))
        options["resolution"] = 1000
        options["geofile"] = os.path.join(options["dir"], options["geofile"])
        options.update(kwargs)

        fparser = Parser(options.get("filename"))
        gparser = Parser(options.get("geofile"))

        if isinstance(filename, (list, set, tuple)):
            # we got the entire dataset.
            for fname in filename:
                if fnmatch(os.path.basename(fname), fparser.globify()):
                    metadata = fparser.parse(os.path.basename(fname))
                    resolution = self.res[metadata["resolution"]]
                    self.datafiles[resolution] = fname
                elif fnmatch(os.path.basename(fname), gparser.globify()):
                    self.geofile = fname
        elif ((filename is not None) and
              fnmatch(os.path.basename(options["filename"]), fparser.globify())):
            # read just one file
            logger.debug("Reading from file: " + str(options["filename"]))
            filename = options["filename"]
            resolution = self.res[os.path.basename(filename)[5]]
            self.datafiles[resolution] = filename
        if not self.datafiles:
            # find files according to config
            logger.debug(
                "Didn't get any valid file as input, looking in defined places")
            resolution = int(options["resolution"]) or 1000

            for res in [250, 500, 1000]:
                datafile = globify(os.path.join(options['dir'],
                                                options["filename"]),
                                   {'resolution': self.inv_res[res],
                                    'start_time': satscene.time_slot})
                try:
                    self.datafiles[res] = check_filename(datafile)
                except IOError:
                    self.datafiles[res] = None
                    logger.warning("Can't find file for resolution %s with template: %s",
                                   str(res), datafile)

            try:
                self.geofile = check_filename(globify(options["geofile"],
                                                      {'start_time': satscene.time_slot}))
            except IOError:
                self.geofile = None
                logger.warning("Can't find geofile with template: %s",
                               options['geofile'])

        resolution = options["resolution"]
        cores = options.get("cores", max(multiprocessing.cpu_count() / 4, 1))

        datadict = {
            1000: ['EV_250_Aggr1km_RefSB',
                   'EV_500_Aggr1km_RefSB',
                   'EV_1KM_RefSB',
                   'EV_1KM_Emissive'],
            500: ['EV_250_Aggr500_RefSB',
                  'EV_500_RefSB'],
            250: ['EV_250_RefSB']}

        loaded_bands = []

        # process by dataset, reflective and emissive datasets separately

        resolutions = [250, 500, 1000]

        for res in resolutions:
            if res < resolution:
                continue
            logger.debug("Working on resolution %d", res)
            self.filename = self.datafiles[res]

            logger.debug("Using " + str(cores) + " cores for interpolation")

            try:
                self.data = SD(str(self.filename))
            except HDF4Error as err:
                logger.warning("Could not load data from " + str(self.filename)
                               + ": " + str(err))
                continue

            datasets = datadict[res]
            for dataset in datasets:
                subdata = self.data.select(dataset)
                band_names = subdata.attributes()["band_names"].split(",")
                if len(satscene.channels_to_load & set(band_names)) > 0:
                    # get the relative indices of the desired channels
                    indices = [i for i, band in enumerate(band_names)
                               if band in satscene.channels_to_load]
                    uncertainty = self.data.select(dataset + "_Uncert_Indexes")
                    if dataset.endswith('Emissive'):
                        array = calibrate_tb(
                            subdata, uncertainty, indices, band_names)
                    else:
                        array = calibrate_refl(subdata, uncertainty, indices)
                    for (i, idx) in enumerate(indices):
                        if band_names[idx] in loaded_bands:
                            continue
                        satscene[band_names[idx]] = array[i]
                        # fix the resolution to match the loaded data.
                        satscene[band_names[idx]].resolution = res
                        loaded_bands.append(band_names[idx])

        # Get the orbit number
        if not satscene.orbit:
            mda = self.data.attributes()["CoreMetadata.0"]
            orbit_idx = mda.index("ORBITNUMBER")
            satscene.orbit = int(mda[orbit_idx + 111:orbit_idx + 116])

        # Get the geolocation
        # if resolution != 1000:
        #    logger.warning("Cannot load geolocation at this resolution (yet).")
        #    return

        for band_name in loaded_bands:
            lon, lat = self.get_lonlat(
                satscene[band_name].resolution, satscene.time_slot, cores)
            area = geometry.SwathDefinition(lons=lon, lats=lat)
            satscene[band_name].area = area

        # Trimming out dead sensor lines (detectors) on aqua:
        # (in addition channel 21 is noisy)
        if satscene.satname == "aqua":
            for band in ["6", "27", "36"]:
                if not satscene[band].is_loaded() or satscene[band].data.mask.all():
                    continue
                width = satscene[band].data.shape[1]
                height = satscene[band].data.shape[0]
                indices = satscene[band].data.mask.sum(1) < width
                if indices.sum() == height:
                    continue
                satscene[band] = satscene[band].data[indices, :]
                satscene[band].area = geometry.SwathDefinition(
                    lons=satscene[band].area.lons[indices, :],
                    lats=satscene[band].area.lats[indices, :])

        # Trimming out dead sensor lines (detectors) on terra:
        # (in addition channel 27, 30, 34, 35, and 36 are nosiy)
        if satscene.satname == "terra":
            for band in ["29"]:
                if not satscene[band].is_loaded() or satscene[band].data.mask.all():
                    continue
                width = satscene[band].data.shape[1]
                height = satscene[band].data.shape[0]
                indices = satscene[band].data.mask.sum(1) < width
                if indices.sum() == height:
                    continue
                satscene[band] = satscene[band].data[indices, :]
                satscene[band].area = geometry.SwathDefinition(
                    lons=satscene[band].area.lons[indices, :],
                    lats=satscene[band].area.lats[indices, :])

        for band_name in loaded_bands:
            band_uid = hashlib.sha1(satscene[band_name].data.mask).hexdigest()
            satscene[band_name].area.area_id = ("swath_" + satscene.fullname + "_"
                                                + str(satscene.time_slot) + "_"
                                                +
                                                str(satscene[
                                                    band_name].shape) + "_"
                                                + str(band_uid))
            satscene[band_name].area_id = satscene[band_name].area.area_id

    def get_lonlat(self, resolution, time_slot, cores=1):
        """Read lat and lon.
        """
        if (resolution, time_slot) in self.areas:
            return self.areas[resolution, time_slot]
        logger.debug("generating lon, lat at %d", resolution)
        if self.geofile is not None:
            coarse_resolution = 1000
            filename = self.geofile
        else:
            coarse_resolution = 5000
            logger.info("Using 5km geolocation and interpolating")
            filename = (self.datafiles.get(1000) or
                        self.datafiles.get(500) or
                        self.datafiles.get(250))

        logger.debug("Loading geolocation from file: " + str(filename)
                     + " at resolution " + str(coarse_resolution))

        data = SD(str(filename))
        lat = data.select("Latitude")
        fill_value = lat.attributes()["_FillValue"]
        lat = np.ma.masked_equal(lat.get(), fill_value)
        lon = data.select("Longitude")
        fill_value = lon.attributes()["_FillValue"]
        lon = np.ma.masked_equal(lon.get(), fill_value)

        if resolution == coarse_resolution:
            self.areas[resolution] = lon, lat
            return lon, lat

        from geotiepoints import modis5kmto1km, modis1kmto500m, modis1kmto250m
        logger.debug("Interpolating from " + str(coarse_resolution)
                     + " to " + str(resolution))
        if coarse_resolution == 5000:
            lon, lat = modis5kmto1km(lon, lat)
        if resolution == 500:
            lon, lat = modis1kmto500m(lon, lat, cores)
        if resolution == 250:
            lon, lat = modis1kmto250m(lon, lat, cores)

        self.areas[resolution, time_slot] = lon, lat
        return lon, lat

    # These have to be interpolated...
    def get_height(self):
        return self.data.select("Height")

    def get_sunz(self):
        return self.data.select("SolarZenith")

    def get_suna(self):
        return self.data.select("SolarAzimuth")

    def get_satz(self):
        return self.data.select("SensorZenith")

    def get_sata(self):
        return self.data.select("SensorAzimuth")


def load(satscene, *args, **kwargs):
    """Read data from file and load it into *satscene*.
    """
    del args
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = kwargs
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value
    options["resolution"] = kwargs.get("resolution", 1000)
    options["filename"] = kwargs.get("filename")

    CASES[satscene.instrument_name](satscene, options)


def calibrate_refl(subdata, uncertainty, indices):
    """Calibration for reflective channels.
    """
    del uncertainty
    #uncertainty_array = uncertainty.get()
    # array = np.ma.MaskedArray(subdata.get(),
    #                          mask=(uncertainty_array >= 15))

    # FIXME: The loading should not be done here.

    array = np.vstack(np.expand_dims(subdata[idx, :, :], 0) for idx in indices)
    valid_range = subdata.attributes()["valid_range"]
    array = np.ma.masked_outside(array,
                                 valid_range[0],
                                 valid_range[1],
                                 copy=False)
    array = array * np.float32(1.0)
    offsets = np.array(subdata.attributes()["reflectance_offsets"],
                       dtype=np.float32)[indices]
    scales = np.array(subdata.attributes()["reflectance_scales"],
                      dtype=np.float32)[indices]
    dims = (len(indices), 1, 1)
    array = (array - offsets.reshape(dims)) * scales.reshape(dims) * 100
    return array


def calibrate_tb(subdata, uncertainty, indices, band_names):
    """Calibration for the emissive channels.
    """
    del uncertainty
    #uncertainty_array = uncertainty.get()
    # array = np.ma.MaskedArray(subdata.get(),
    #                          mask=(uncertainty_array >= 15))

    # FIXME: The loading should not be done here.

    array = np.vstack(np.expand_dims(subdata[idx, :, :], 0) for idx in indices)
    valid_range = subdata.attributes()["valid_range"]
    array = np.ma.masked_outside(array,
                                 valid_range[0],
                                 valid_range[1],
                                 copy=False)

    offsets = np.array(subdata.attributes()["radiance_offsets"],
                       dtype=np.float32)[indices]
    scales = np.array(subdata.attributes()["radiance_scales"],
                      dtype=np.float32)[indices]

    #- Planck constant (Joule second)
    h__ = np.float32(6.6260755e-34)

    #- Speed of light in vacuum (meters per second)
    c__ = np.float32(2.9979246e+8)

    #- Boltzmann constant (Joules per Kelvin)
    k__ = np.float32(1.380658e-23)

    #- Derived constants
    c_1 = 2 * h__ * c__ * c__
    c_2 = (h__ * c__) / k__

    #- Effective central wavenumber (inverse centimeters)
    cwn = np.array([
        2.641775E+3, 2.505277E+3, 2.518028E+3, 2.465428E+3,
        2.235815E+3, 2.200346E+3, 1.477967E+3, 1.362737E+3,
        1.173190E+3, 1.027715E+3, 9.080884E+2, 8.315399E+2,
        7.483394E+2, 7.308963E+2, 7.188681E+2, 7.045367E+2],
        dtype=np.float32)

    #- Temperature correction slope (no units)
    tcs = np.array([
        9.993411E-1, 9.998646E-1, 9.998584E-1, 9.998682E-1,
        9.998819E-1, 9.998845E-1, 9.994877E-1, 9.994918E-1,
        9.995495E-1, 9.997398E-1, 9.995608E-1, 9.997256E-1,
        9.999160E-1, 9.999167E-1, 9.999191E-1, 9.999281E-1],
        dtype=np.float32)

    #- Temperature correction intercept (Kelvin)
    tci = np.array([
        4.770532E-1, 9.262664E-2, 9.757996E-2, 8.929242E-2,
        7.310901E-2, 7.060415E-2, 2.204921E-1, 2.046087E-1,
        1.599191E-1, 8.253401E-2, 1.302699E-1, 7.181833E-2,
        1.972608E-2, 1.913568E-2, 1.817817E-2, 1.583042E-2],
        dtype=np.float32)

    # Transfer wavenumber [cm^(-1)] to wavelength [m]
    cwn = 1 / (cwn * 100)

    # Some versions of the modis files do not contain all the bands.
    emmissive_channels = ["20", "21", "22", "23", "24", "25", "27", "28", "29",
                          "30", "31", "32", "33", "34", "35", "36"]
    current_channels = [i for i, band in enumerate(emmissive_channels)
                        if band in band_names]
    global_indices = list(np.array(current_channels)[indices])

    dims = (len(indices), 1, 1)
    cwn = cwn[global_indices].reshape(dims)
    tcs = tcs[global_indices].reshape(dims)
    tci = tci[global_indices].reshape(dims)

    tmp = (array - offsets.reshape(dims)) * scales.reshape(dims)
    tmp = c_2 / (cwn * np.ma.log(c_1 / (1000000 * tmp * cwn ** 5) + 1))
    array = (tmp - tci) / tcs
    return array


def load_modis(satscene, options):
    """Read modis data from file and load it into *satscene*.

    *resolution* parameters specifies in which resolution to load the data. If
     the specified resolution is not available for the channel, it is NOT
     loaded. If no resolution is specified, the 1km resolution (aggregated) is
     used.
    """
    if options["filename"] is not None:
        logger.debug("Reading from file: " + str(options["filename"]))
        filename = options["filename"]
        res = {"1": 1000,
               "Q": 250,
               "H": 500}
        resolution = res[os.path.split(filename)[1][5]]
    else:
        resolution = int(options["resolution"]) or 1000

        filename_tmpl = satscene.time_slot.strftime(options["filename" +
                                                            str(resolution)])
        file_list = glob.glob(os.path.join(options["dir"], filename_tmpl))
        if len(file_list) > 1:
            raise IOError("More than 1 file matching!")
        elif len(file_list) == 0:
            raise IOError("No EOS MODIS file matching " +
                          filename_tmpl + " in " +
                          options["dir"])
        filename = file_list[0]

    cores = options.get("cores", 1)

    logger.debug("Using " + str(cores) + " cores for interpolation")

    load_generic(satscene, filename, resolution, cores)


def load_generic(satscene, filename, resolution, cores):
    """Read modis data, generic part.
    """

    try:
        data = SD(str(filename))
    except HDF4Error as err:
        logger.warning("Could not load data from " + str(filename)
                       + ": " + str(err))
        return

    datadict = {
        1000: ['EV_250_Aggr1km_RefSB',
               'EV_500_Aggr1km_RefSB',
               'EV_1KM_RefSB',
               'EV_1KM_Emissive'],
        500: ['EV_250_Aggr500_RefSB',
              'EV_500_RefSB'],
        250: ['EV_250_RefSB']}

    datasets = datadict[resolution]

    loaded_bands = []

    # process by dataset, reflective and emissive datasets separately

    for dataset in datasets:
        subdata = data.select(dataset)
        band_names = subdata.attributes()["band_names"].split(",")
        if len(satscene.channels_to_load & set(band_names)) > 0:
            # get the relative indices of the desired channels
            indices = [i for i, band in enumerate(band_names)
                       if band in satscene.channels_to_load]
            uncertainty = data.select(dataset + "_Uncert_Indexes")
            if dataset.endswith('Emissive'):
                array = calibrate_tb(subdata, uncertainty, indices, band_names)
            else:
                array = calibrate_refl(subdata, uncertainty, indices)
            for (i, idx) in enumerate(indices):
                satscene[band_names[idx]] = array[i]
                # fix the resolution to match the loaded data.
                satscene[band_names[idx]].resolution = resolution
                loaded_bands.append(band_names[idx])

    # Get the orbit number
    if not satscene.orbit:
        mda = data.attributes()["CoreMetadata.0"]
        orbit_idx = mda.index("ORBITNUMBER")
        satscene.orbit = int(mda[orbit_idx + 111:orbit_idx + 116])

    # Get the geolocation
    # if resolution != 1000:
    #    logger.warning("Cannot load geolocation at this resolution (yet).")
    #    return

    lat, lon = get_lat_lon(satscene, resolution, filename, cores)
    area = geometry.SwathDefinition(lons=lon, lats=lat)
    for band_name in loaded_bands:
        satscene[band_name].area = area

    # Trimming out dead sensor lines (detectors) on aqua:
    # (in addition channel 21 is noisy)
    if satscene.satname == "aqua":
        for band in ["6", "27", "36"]:
            if not satscene[band].is_loaded() or satscene[band].data.mask.all():
                continue
            width = satscene[band].data.shape[1]
            height = satscene[band].data.shape[0]
            indices = satscene[band].data.mask.sum(1) < width
            if indices.sum() == height:
                continue
            satscene[band] = satscene[band].data[indices, :]
            satscene[band].area = geometry.SwathDefinition(
                lons=satscene[band].area.lons[indices, :],
                lats=satscene[band].area.lats[indices, :])
            satscene[band].area.area_id = ("swath_" + satscene.fullname + "_"
                                           + str(satscene.time_slot) + "_"
                                           + str(satscene[band].shape) + "_"
                                           + str(band))

    # Trimming out dead sensor lines (detectors) on terra:
    # (in addition channel 27, 30, 34, 35, and 36 are nosiy)
    if satscene.satname == "terra":
        for band in ["29"]:
            if not satscene[band].is_loaded() or satscene[band].data.mask.all():
                continue
            width = satscene[band].data.shape[1]
            height = satscene[band].data.shape[0]
            indices = satscene[band].data.mask.sum(1) < width
            if indices.sum() == height:
                continue
            satscene[band] = satscene[band].data[indices, :]
            satscene[band].area = geometry.SwathDefinition(
                lons=satscene[band].area.lons[indices, :],
                lats=satscene[band].area.lats[indices, :])
            satscene[band].area.area_id = ("swath_" + satscene.fullname + "_"
                                           + str(satscene.time_slot) + "_"
                                           + str(satscene[band].shape) + "_"
                                           + str(band))


def get_lat_lon(satscene, resolution, filename, cores=1):
    """Read lat and lon.
    """

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value

    options["filename"] = filename
    options["resolution"] = resolution
    options["cores"] = cores
    return LAT_LON_CASES[satscene.instrument_name](satscene, options)


def get_lat_lon_modis(satscene, options):
    """Read lat and lon.
    """
    filename_tmpl = satscene.time_slot.strftime(options["geofile"])
    file_list = glob.glob(os.path.join(options["dir"], filename_tmpl))

    if len(file_list) == 0:
        # Try in the same directory as the data
        data_dir = os.path.split(options["filename"])[0]
        file_list = glob.glob(os.path.join(data_dir, filename_tmpl))

    if len(file_list) > 1:
        logger.warning("More than 1 geolocation file matching!")
        filename = max(file_list, key=lambda x: os.stat(x).st_mtime)
        coarse_resolution = 1000
    elif len(file_list) == 0:
        logger.warning("No geolocation file matching " + filename_tmpl
                       + " in " + options["dir"])
        logger.debug("Using 5km geolocation and interpolating")
        filename = options["filename"]
        coarse_resolution = 5000
    else:
        filename = file_list[0]
        coarse_resolution = 1000

    logger.debug("Loading geolocation file: " + str(filename)
                 + " at resolution " + str(coarse_resolution))

    resolution = options["resolution"]

    data = SD(str(filename))
    lat = data.select("Latitude")
    fill_value = lat.attributes()["_FillValue"]
    lat = np.ma.masked_equal(lat.get(), fill_value)
    lon = data.select("Longitude")
    fill_value = lon.attributes()["_FillValue"]
    lon = np.ma.masked_equal(lon.get(), fill_value)

    if resolution == coarse_resolution:
        return lat, lon

    cores = options["cores"]

    from geotiepoints import modis5kmto1km, modis1kmto500m, modis1kmto250m
    logger.debug("Interpolating from " + str(coarse_resolution)
                 + " to " + str(resolution))
    if coarse_resolution == 5000:
        lon, lat = modis5kmto1km(lon, lat)
    if resolution == 500:
        lon, lat = modis1kmto500m(lon, lat, cores)
    if resolution == 250:
        lon, lat = modis1kmto250m(lon, lat, cores)

    return lat, lon


def get_lonlat(satscene, row, col):
    """Estimate lon and lat.
    """
    import glob

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    path = conf.get("modis-level2", "dir")
    geofile_tmpl = conf.get("modis-level2", "geofile")

    filename_tmpl = satscene.time_slot.strftime(geofile_tmpl)
    file_list = glob.glob(os.path.join(path, filename_tmpl))

    if len(file_list) > 1:
        raise IOError("More than 1 geolocation file matching!" + filename_tmpl)
    elif len(file_list) == 0:
        logger.info("No MODIS geolocation file matching: " + filename_tmpl
                    + ", estimating")
        filename = ""
    else:
        filename = file_list[0]
        logger.debug("Geolocation file = " + filename)

    if(os.path.exists(filename) and
       (satscene.lon is None or satscene.lat is None)):
        data = SD(filename)
        lat = data.select("Latitude")
        fill_value = lat.attributes()["_FillValue"]
        satscene.lat = np.ma.masked_equal(lat.get(), fill_value)
        lon = data.select("Longitude")
        fill_value = lon.attributes()["_FillValue"]
        satscene.lon = np.ma.masked_equal(lon.get(), fill_value)

    estimate = True

    try:
        lon = satscene.lon[row, col]
        lat = satscene.lat[row, col]
        if(satscene.lon.mask[row, col] == False and
           satscene.lat.mask[row, col] == False):
            estimate = False
    except TypeError:
        pass
    except IndexError:
        pass

    if not estimate:
        return lon, lat

    from mpop.saturn.two_line_elements import Tle

    tle = Tle(satellite=satscene.satname)
    track_start = tle.get_latlonalt(satscene.time_slot)
    track_end = tle.get_latlonalt(satscene.time_slot + satscene.granularity)

    # WGS84
    # flattening
    f__ = 1 / 298.257223563
    # semi_major_axis
    a__ = 6378137.0

    s__, alpha12, alpha21 = vinc_dist(f__, a__,
                                      track_start[0], track_start[1],
                                      track_end[0], track_end[1])
    scanlines = satscene.granularity.seconds / satscene.span

    if row < scanlines / 2:
        if row == 0:
            track_now = track_start
        else:
            track_now = vinc_pt(f__, a__, track_start[0], track_start[1],
                                alpha12, (s__ * row) / scanlines)
        lat_now = track_now[0]
        lon_now = track_now[1]

        s__, alpha12, alpha21 = vinc_dist(f__, a__,
                                          lat_now, lon_now,
                                          track_end[0], track_end[1])
        fac = 1
    else:
        if scanlines - row - 1 == 0:
            track_now = track_end
        else:
            track_now = vinc_pt(f__, a__, track_end[0], track_end[1], alpha21,
                                (s__ * (scanlines - row - 1)) / scanlines)
        lat_now = track_now[0]
        lon_now = track_now[1]

        s__, alpha12, alpha21 = vinc_dist(f__, a__,
                                          lat_now, lon_now,
                                          track_start[0], track_start[1])
        fac = -1

    if col < 1354 / 2:
        lat, lon, alp = vinc_pt(f__, a__, lat_now, lon_now,
                                alpha12 + np.pi / 2 * fac,
                                2340000.0 / 2 - (2340000.0 / 1354) * col)
    else:
        lat, lon, alp = vinc_pt(f__, a__, lat_now, lon_now,
                                alpha12 - np.pi / 2 * fac,
                                (2340000.0 / 1354) * col - 2340000.0 / 2)
        del alp
    lon = np.rad2deg(lon)
    lat = np.rad2deg(lat)

    if lon > 180:
        lon -= 360
    if lon <= -180:
        lon += 360

    return lon, lat


def vinc_dist(f__, a__, phi1, lembda1, phi2, lembda2):
    """ 

    Returns the distance between two geographic points on the ellipsoid
    and the forward and reverse azimuths between these points.
    lats, longs and azimuths are in radians, distance in metres 

    Returns ( s__, alpha12,  alpha21 ) as a tuple

    """

    if (abs(phi2 - phi1) < 1e-8) and (abs(lembda2 - lembda1) < 1e-8):
        return 0.0, 0.0, 0.0

    two_pi = 2.0 * math.pi

    b__ = a__ * (1.0 - f__)

    tan_u1 = (1 - f__) * math.tan(phi1)
    tan_u2 = (1 - f__) * math.tan(phi2)

    u_1 = math.atan(tan_u1)
    u_2 = math.atan(tan_u2)

    lembda = lembda2 - lembda1
    last_lembda = -4000000.0                # an impossibe value
    omega = lembda

    # Iterate the following equations,
    #  until there is no significant change in lembda

    while (last_lembda < -3000000.0 or
           lembda != 0 and
           abs((last_lembda - lembda) / lembda) > 1.0e-9):

        sqr_sin_sigma = (pow(math.cos(u_2) * math.sin(lembda), 2) +
                         pow((math.cos(u_1) * math.sin(u_2) -
                              math.sin(u_1) * math.cos(u_2) *
                              math.cos(lembda)), 2))

        sin_sigma = math.sqrt(sqr_sin_sigma)

        cos_sigma = (math.sin(u_1) * math.sin(u_2) +
                     math.cos(u_1) * math.cos(u_2) * math.cos(lembda))

        sigma = math.atan2(sin_sigma, cos_sigma)

        sin_alpha = (math.cos(u_1) * math.cos(u_2) * math.sin(lembda) /
                     math.sin(sigma))
        alpha = math.asin(sin_alpha)

        cos2sigma_m = (math.cos(sigma) -
                       (2 * math.sin(u_1) * math.sin(u_2) /
                        pow(math.cos(alpha), 2)))

        c__ = ((f__ / 16) * pow(math.cos(alpha), 2) *
               (4 + f__ * (4 - 3 * pow(math.cos(alpha), 2))))

        last_lembda = lembda

        lembda = (omega + (1 - c__) * f__ * math.sin(alpha) *
                  (sigma + c__ * math.sin(sigma) *
                   (cos2sigma_m + c__ * math.cos(sigma) *
                    (-1 + 2 * pow(cos2sigma_m, 2)))))

    u2_ = pow(math.cos(alpha), 2) * (a__ * a__ - b__ * b__) / (b__ * b__)

    aa_ = 1 + (u2_ / 16384) * (4096 + u2_ * (-768 + u2_ * (320 - 175 * u2_)))

    bb_ = (u2_ / 1024) * (256 + u2_ * (-128 + u2_ * (74 - 47 * u2_)))

    delta_sigma = bb_ * sin_sigma * (cos2sigma_m + (bb_ / 4) *
                                     (cos_sigma * (-1 + 2 * pow(cos2sigma_m, 2)) -
                                      (bb_ / 6) * cos2sigma_m * (-3 + 4 * sqr_sin_sigma) *
                                      (-3 + 4 * pow(cos2sigma_m, 2))))

    s__ = b__ * aa_ * (sigma - delta_sigma)

    alpha12 = (math.atan2((math.cos(u_2) * math.sin(lembda)),
                          (math.cos(u_1) * math.sin(u_2) -
                           math.sin(u_1) * math.cos(u_2) * math.cos(lembda))))

    alpha21 = (math.atan2((math.cos(u_1) * math.sin(lembda)),
                          (-math.sin(u_1) * math.cos(u_2) +
                           math.cos(u_1) * math.sin(u_2) * math.cos(lembda))))

    if (alpha12 < 0.0):
        alpha12 = alpha12 + two_pi
    if (alpha12 > two_pi):
        alpha12 = alpha12 - two_pi

    alpha21 = alpha21 + two_pi / 2.0
    if (alpha21 < 0.0):
        alpha21 = alpha21 + two_pi
    if (alpha21 > two_pi):
        alpha21 = alpha21 - two_pi

    return s__, alpha12,  alpha21

# END of Vincenty's Inverse formulae


#----------------------------------------------------------------------------
# Vincenty's Direct formulae                                                |
# Given: latitude and longitude of a point (phi1, lembda1) and              |
# the geodetic azimuth (alpha12)                                            |
# and ellipsoidal distance in metres (s) to a second point,                 |
#                                                                           |
# Calculate: the latitude and longitude of the second point (phi2, lembda2) |
# and the reverse azimuth (alpha21).                                        |
#                                                                           |
#----------------------------------------------------------------------------

def vinc_pt(f__, a__, phi1, lembda1, alpha12, s__):
    """

    Returns the lat and long of projected point and reverse azimuth
    given a reference point and a distance and azimuth to project.
    lats, longs and azimuths are passed in decimal degrees

    Returns ( phi2,  lambda2,  alpha21 ) as a tuple 

    """

    two_pi = 2.0 * math.pi

    if (alpha12 < 0.0):
        alpha12 = alpha12 + two_pi
    if (alpha12 > two_pi):
        alpha12 = alpha12 - two_pi

    b__ = a__ * (1.0 - f__)

    tan_u1 = (1 - f__) * math.tan(phi1)
    u_1 = math.atan(tan_u1)
    sigma1 = math.atan2(tan_u1, math.cos(alpha12))
    sinalpha = math.cos(u_1) * math.sin(alpha12)
    cosalpha_sq = 1.0 - sinalpha * sinalpha

    u_2 = cosalpha_sq * (a__ * a__ - b__ * b__) / (b__ * b__)
    aa_ = 1.0 + (u_2 / 16384) * (4096 + u_2 * (-768 + u_2 *
                                               (320 - 175 * u_2)))
    bb_ = (u_2 / 1024) * (256 + u_2 * (-128 + u_2 * (74 - 47 * u_2)))

    # Starting with the approximation
    sigma = (s__ / (b__ * aa_))

    last_sigma = 2.0 * sigma + 2.0  # something impossible

    # Iterate the following three equations
    # until there is no significant change in sigma

    # two_sigma_m , delta_sigma

    while (abs((last_sigma - sigma) / sigma) > 1.0e-9):

        two_sigma_m = 2 * sigma1 + sigma

        delta_sigma = (bb_ * math.sin(sigma) *
                       (math.cos(two_sigma_m)
                        + (bb_ / 4) * (math.cos(sigma) *
                                       (-1 + 2 * math.pow(math.cos(two_sigma_m),
                                                          2)
                                        - (bb_ / 6) * math.cos(two_sigma_m) *
                                        (-3 + 4 * math.pow(math.sin(sigma), 2)) *
                                        (-3 + 4 * math.pow(math.cos(two_sigma_m),
                                                           2))))))
        last_sigma = sigma
        sigma = (s__ / (b__ * aa_)) + delta_sigma

    phi2 = math.atan2((math.sin(u_1) * math.cos(sigma) +
                       math.cos(u_1) * math.sin(sigma) * math.cos(alpha12)),
                      ((1 - f__) * math.sqrt(math.pow(sinalpha, 2) +
                                             pow(math.sin(u_1) *
                                                 math.sin(sigma) -
                                                 math.cos(u_1) *
                                                 math.cos(sigma) *
                                                 math.cos(alpha12), 2))))

    lembda = math.atan2((math.sin(sigma) * math.sin(alpha12)),
                        (math.cos(u_1) * math.cos(sigma) -
                         math.sin(u_1) * math.sin(sigma) * math.cos(alpha12)))

    cc_ = (f__ / 16) * cosalpha_sq * (4 + f__ * (4 - 3 * cosalpha_sq))

    omega = lembda - (1 - cc_) * f__ * sinalpha *  \
        (sigma + cc_ * math.sin(sigma) * (math.cos(two_sigma_m) +
                                          cc_ * math.cos(sigma) *
                                          (-1 + 2 *
                                           math.pow(math.cos(two_sigma_m), 2))))

    lembda2 = lembda1 + omega

    alpha21 = math.atan2(sinalpha, (-math.sin(u_1) * math.sin(sigma) +
                                    math.cos(u_1) * math.cos(sigma) *
                                    math.cos(alpha12)))

    alpha21 = alpha21 + two_pi / 2.0
    if (alpha21 < 0.0):
        alpha21 = alpha21 + two_pi
    if (alpha21 > two_pi):
        alpha21 = alpha21 - two_pi

    return phi2,  lembda2,  alpha21

# END of Vincenty's Direct formulae


CASES = {
    "modis": load_modis
}

LAT_LON_CASES = {
    "modis": get_lat_lon_modis
}


if __name__ == "__main__":
    filenames = [u'/data/prod/satellit/modis/lvl1/thin_MYD021KM.A2015287.0255.005.2015287051016.NRT.hdf',
                 u'/data/prod/satellit/modis/lvl1/thin_MYD021KM.A2015287.0300.005.2015287050819.NRT.hdf',
                 u'/data/prod/satellit/modis/lvl1/thin_MYD021KM.A2015287.0305.005.2015287050825.NRT.hdf']

    from mpop.utils import debug_on
    debug_on()
    from mpop.satellites import PolarFactory
    from datetime import datetime
    time_slot = datetime(2015, 10, 14, 2, 55)
    orbit = "18181"
    global_data = PolarFactory.create_scene(
        "EARSEOS-Aqua", "", "modis", time_slot, orbit)

    global_data.load([3.75, 0.555, 0.551, 7.3, 1.63, 10.8, 0.488, 12.0, 0.85, 0.469, 0.748, 0.443, 0.645, 6.7, 0.635,
                      8.7, 0.412], filename=filenames)

    #global_data.channels_to_load = set(['31'])
    #reader = ModisReader(global_data)
    #reader.load(global_data, filename=filenames)
    print global_data
    # global_data[10.8].show()



"""
Reader for EUMETSATs Hydrology SAF (HSAF) h03 product
HSAF website http://hsaf.meteoam.it
h03 product is precipitation rate at the ground 
by GEO(MSG)/Infrared supported by LEO/Microwave
http://hsaf.meteoam.it/precipitation.php?tab=3

After registration the data is available from 
ftp://ftphsaf.meteoam.it/h03

possible accepted formats for this reader are:
* grib as provided by HSAF
* netCDF (grib file converted with cdo)

- Initial version: 
  2015-07-23 Ulrich Hamann (MeteoSwiss)
"""

from ConfigParser import ConfigParser
from mpop import CONFIG_PATH
import os
import numpy.ma as ma
from glob import glob
import datetime

def load(satscene, **kargs):
    """Reader for EUMETSATs Hydrology SAF (HSAF) h03 product
    h03 product is precipitation rate at the ground 
    by GEO(MSG)/Infrared supported by LEO/Microwave
    http://hsaf.meteoam.it/precipitation.php?tab=3
    """

    # Read config file content
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    values = {"orbit": satscene.orbit,
    "satname": satscene.satname,
    "number": satscene.number,
    "instrument": satscene.instrument_name,
    "satellite": satscene.fullname
    }

    # end of scan time 12min after start 
    end_time = satscene.time_slot + datetime.timedelta(minutes=12)

    filepath    = end_time.strftime(conf.get("seviri-level2", "dir",raw=True))
    filepattern = end_time.strftime(conf.get("seviri-level2", "filename",raw=True)) % values
    filename = os.path.join( filepath, filepattern)

    print "... search for file: ", filename
    filenames=glob(str(filename))
    if len(filenames) == 0:
        print "*** Error, no file found"
        quit()
    elif len(filenames) > 1:
        print "*** Warning, more than 1 datafile found: "
        for filename in filenames:
            print "    ", filename
        
    # possible formats: h03_20150513_1557_rom.grb.gz, h03_20150513_1612_rom.grb, h03_20150513_1612_rom.nc
    fileformats = [filename.split(".")[-1] for filename in filenames]
    # try to find grb file 

    if 'grb' in fileformats:
        # read grib 
        data, fill_value, units, long_name = read_h03_grib(filenames[fileformats.index('grb')])
    elif 'nc' in fileformats:
        # read netCDF
        data, fill_value, units, long_name = read_h03_netCDF(filenames[fileformats.index('nc')])
    elif 'gz' in fileformats:
        # unzip 
        from subprocess import call
        infile = filenames[fileformats.index('gz')]
        outfile = infile[:-3]
        print "    unizp ", infile 
        # gunzip -c h03_20150513_1557_rom.grb.gz > h03_20150513_1557_rom.grb
        # call("/bin/gunzip "+ infile                +" 2>&1", shell=True) # dont keep gz file 
        call("/bin/gunzip -c "+ infile+" > "+ outfile  +" 2>&1", shell=True) # keep gz file 
        # check format of gunziped file
        if outfile.split(".")[-1] == 'grb':
            data, fill_value, units, long_name = read_h03_grib(outfile)
        elif outfile.split(".")[-1] == 'nc':
            data, fill_value, units, long_name = read_h03_netCDF(outfile)

    if units == "kg m**-2 s**-1" or units == "kg m-2s-1":
        data *= 3600 
        units = "kg m-2 h-1"

    satscene['h03'] = data
    satscene['h03'].fill_value = fill_value
    satscene['h03'].units      = units
    satscene['h03'].long_name  = long_name
    satscene['h03'].product_name = 'h03'

    # personal communication with help desk
    # Each H03 grib file contains precipitation data of a 900x1900 pixel sub-area of the SEVIRI full disk area (3712x3712 pixels). 
    # The first pixel  of H03 (pixel (1,1)) grib file corresponds to Seviri pixel (1095,85) if the Seviri pixel (1,1) is in the Nort-East. 
    # I can confirm that only the prime satellite is used (position subsatellite longitude 0 degree East).
    # For the future we are thinking to disseminate the h03 outputs already corrected in parallax.

    # conversion of above information to correct AreaDefinition 
    # full_disk = get_area_def("SeviriDiskFull")
    # from mpop.projector import get_area_def
    # import numpy as np
    # np.array(area_def.get_proj_coords(data_slice=(85+900,1095     ))) - 3000.40316582 / 2.
    #    array([-2284807.01076965,  2611850.9558437 ])
    # np.array(area_def.get_proj_coords(data_slice=(85    ,1095+1900))) + 3000.40316582 / 2.
    #                                        array([ 3418959.40744847,  5315214.20824482])
    # or 
    # aex = full_disk.get_area_extent_for_subsets(985,1095,85,2995)

    proj = {'proj': 'geos', 'a': '6378169.0', 'b': '6356583.8', 'h': '35785831.0', 'lon_0': '0.0'}
    aex =      (-2284807.01076965, 2611850.9558437,  3418959.40744847,  5315214.20824482)

    from pyresample.geometry import AreaDefinition
    satscene.area = AreaDefinition("hsaf",
                                   "hsaf",
                                   "geos0",
                                   proj,
                                   1900,
                                   900,
                                   aex)

def read_h03_grib(filename):

    try:
        import pygrib
    except ImportError:
        print "... module pygrib needs to be installed"
        quit()

    # see http://pygrib.googlecode.com/svn/trunk/docs/pygrib-module.html
    print("... read data from %s" % str(filename))

    grbs = pygrib.open(filename)

    #print(grbs)

    #print 'inventory'
    #for grb in grbs:
    #    print(grb)
    #print 'end inventory'

    long_name  = 'Instantaneous rain rate'
    units      = 'kg m**-2 s**-1'
    _FillValue = 0.0 
    grb = grbs.select(name=long_name)[0]
    # print(grb)
    data = ma.asarray(grb.values)
    data.mask = (data == 0.0) 

    print '    fill_value: ', 0 
    print '    units:      ', units
    print '    long_name:  ', long_name 
    print '    datatype:   ', type(data)
    print '    shape:      ', data.shape
    print '    min/max:    ', data.min(), data.max()

    return data, _FillValue, units, long_name 


def read_h03_netCDF(filename):

    try:
        from netCDF4 import Dataset
    except ImportError:
        print "... module netCDF4 needs to be installed"
        quit()

    print("... read data from %s" % str(filename))

    # Load data from netCDF file
    # see also http://netcdf4-python.googlecode.com/svn/trunk/docs/netCDF4-module.html
    ds = Dataset(filename, 'r')

    if 'irrate' in ds.variables:
        print '    found variable irrate'
        var_name='irrate'            # converted with: cdo -f nc4c -z zip copy infile outfile
    elif 'IRRATE_P30_GSV0' in ds.variables:
        print '    found variable IRRATE_P30_GSV0'
        var_name='IRRATE_P30_GSV0'   # converted with: ncl_convert2nc h03_20150529_0827_rom.grb -e grb -nc4c -cl 9
        #variables:
        #    float IRRATE_P30_GSV0(ygrid_0, xgrid_0) ;
        #            IRRATE_P30_GSV0:initial_time = "05/29/2015 (08:27)" ;
        #            IRRATE_P30_GSV0:parameter_template_discipline_category_number = 30, 3, 1, 1 ;
        #            IRRATE_P30_GSV0:parameter_discipline_and_category = "Space products, Quantitative products" ;
        #            IRRATE_P30_GSV0:grid_type = "Space view perspective or orthographic" ;
        #            IRRATE_P30_GSV0:_FillValue = 1.e+20f ;
        #            IRRATE_P30_GSV0:units = "kg m-2s-1" ;
        #            IRRATE_P30_GSV0:long_name = "Instantaneous rain rate" ;
        #            IRRATE_P30_GSV0:production_status = "Operational test products" ;
        #            IRRATE_P30_GSV0:center = "Rome (RSMC)" ;
        print '*** Error, does not work for unknown reason'
        print '    data.mask = (data == _FillValue) | (data == 0.0) produce error'
        quit()

    #print type(ds.variables[var_name])
    #print dir(ds.variables[var_name])
  
    _FillValue = ds.variables[var_name]._FillValue
    # or fill_value = ds.variables[var_name].getncattr('_FillValue')
    units      = ds.variables[var_name].units
    long_name  = ds.variables[var_name].long_name 

    # Read variable corresponding to channel name
    data = ma.asarray(ds.variables[var_name])

    print '    fill_value: ', ds.variables[var_name]._FillValue 
    print '    units:      ', ds.variables[var_name].units
    print '    long_name:  ', ds.variables[var_name].long_name 
    print '    datatype:   ', ds.variables[var_name].datatype
    print '    shape:      ', data.shape
    print '    min/max:    ', data.min(), data.max()

    if len(data.shape) == 3:
        if data.shape[0] == 1:
            print "   reduce to 2 dimensions (skip time dimension)"
            data = ma.asarray(ds.variables[var_name][0,:,:])
        else:
            print "*** Error, unknown netCDF file format in h03_nc.py"
            print "    probably more time steps in one file (not implemented yet)"
            quit()

    data.mask = (data == _FillValue) | (data == 0.0) 

    return data, _FillValue, units, long_name 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2011 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Minimal script for polar data (from aapp level 1b).

We take the case of level 1b data (calibrated and geolocalized) from noaa 19,
as output from AAPP.

- Install mpop and pyresample
- Don't forget to set up the PPP_CONFIG_DIR variable to point to your
  configuration files.
- Edit the noaa19.cfg configuration file (a template is provided in case
  you don't have one) with your data directory:

  .. code-block:: ini
  
  [avhrr-level2]
  filename = hrpt_%(satellite)s_%Y%m%d_%H%M_%(orbit)s.l1b
  dir = /data/polar/
  format = aapp1b

- Here is an example of a minimal script that has to be called as soon as a new
  swath has arrived

"""

from mpop.satellites import PolarFactory

import sys
from datetime import datetime

if sys.version_info < (2, 5):
    import time
    def strptime(string, fmt=None):
        """This function is available in the datetime module only
        from Python >= 2.5.
        """

        return datetime(*time.strptime(string, fmt)[:6])

else:
    strptime = datetime.strptime

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "Usage: " + sys.argv[0] + " time_string orbit"
        sys.exit()

    time_string = sys.argv[1]
    orbit = sys.argv[2] 
    time_slot = strptime(time_string, "%Y%m%d%H%M")
    global_data = PolarFactory.create_scene("noaa", "19",
                                            "avhrr", time_slot, orbit)

    global_data.load()

    areas = ["euro4", "scan2"]

    for area in areas:

        local_data = global_data.project(area)

        img = local_data.image.overview()
        img.save("overview_" + area + "_" + time_string + ".png")

        img = local_data.image.cloudtop()
        img.save("cloudtop_" + area + "_" + time_string + ".png")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2012, 2013.

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Sara Hornquist <sara.hornquist@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin for reading PPS's cloud products hdf files, in the fileformat used
in PPS v2014.
"""
import ConfigParser
from ConfigParser import NoOptionError

from datetime import datetime, timedelta
import os.path

import mpop.channel
from mpop import CONFIG_PATH
from mpop.utils import get_logger
import numpy as np

import h5py

LOG = get_logger('satin/nwcsaf_pps')

class InfoObject(object):
    """Simple data and info container.
    """
    def __init__(self):
        self.info = {}
        self.data = None

def pack_signed(data, data_type):
    bits = np.iinfo(data_type).bits
    scale_factor = (data.max() - data.min()) / (2**bits - 2)
    add_offset = (data.max() - data.min()) / 2
    no_data = - 2**(bits - 1)
    pack = ((data - add_offset) / scale_factor).astype(data_type)
    return pack, scale_factor, add_offset, no_data


class NwcSafPpsChannel(mpop.channel.GenericChannel):

    def __init__(self, filename=None):
        mpop.channel.GenericChannel.__init__(self)
        self._md = {}
        self._projectables = []
        self._keys = []
        self._refs = {}
        self.shape = None
        if filename:
            self.read(filename)


    def read(self, filename, load_lonlat=True):
        """Read product in hdf format from *filename*
        """
        LOG.debug("Filename: %s" % filename)

        is_temp = False
        if not h5py.is_hdf5(filename):
            # Try see if it is bzipped:
            import bz2
            bz2file = bz2.BZ2File(filename)
            import tempfile
            tmpfilename = tempfile.mktemp()
            try:
                ofpt = open(tmpfilename, 'wb')
                ofpt.write(bz2file.read())
                ofpt.close()
                is_temp = True
            except IOError:
                import traceback
                traceback.print_exc()
                raise IOError("Failed to read the file %s" % filename)

            filename = tmpfilename
            
        if not h5py.is_hdf5(filename):
            if is_temp:
                os.remove(filename)
            raise IOError("File is not a hdf5 file!" % filename)

        h5f = h5py.File(filename, "r")

        # Read the global attributes

        self._md = dict(h5f.attrs)
        self._md["satellite"] = h5f.attrs['platform']
        self._md["orbit"] = h5f.attrs['orbit_number']
        self._md["time_slot"] = (timedelta(seconds=long(h5f.attrs['sec_1970']))
                                 + datetime(1970, 1, 1, 0, 0))

        # Read the data and attributes
        #   This covers only one level of data. This could be made recursive.
        for key, dataset in h5f.iteritems():
            setattr(self, key, InfoObject())
            getattr(self, key).info = dict(dataset.attrs)
            for skey, value in dataset.attrs.iteritems():
                if isinstance(value, h5py.h5r.Reference):
                    self._refs[(key, skey)] = h5f[value].name.split("/")[1]
                    
            if type(dataset.id) is h5py.h5g.GroupID:
                LOG.warning("Format reader does not support groups")
                continue

            try:
                getattr(self, key).data = dataset[:]
                is_palette = (dataset.attrs.get("CLASS", None) == "PALETTE")
                if(len(dataset.shape) > 1 and
                   not is_palette and
                   key not in ["lon", "lat", 
                               "row_indices", "column_indices"]):
                    self._projectables.append(key)
                    if self.shape is None:
                        self.shape = dataset.shape
                    elif self.shape != dataset.shape:
                        raise ValueError("Different variable shapes !")
                else:
                    self._keys.append(key)
            except TypeError:
                setattr(self, key, np.dtype(dataset))
                self._keys.append(key)

        h5f.close()
        
        if is_temp:
            os.remove(filename)

        if not load_lonlat:
            return


        # Setup geolocation
        # We need a no-data mask from one of the projectables to
        # mask out bow-tie deletion pixels from the geolocation array
        # So far only relevant for VIIRS.
        # Preferably the lon-lat data in the PPS VIIRS geolocation
        # file should already be masked. 
        # The no-data values in the products are not only where geo-location is absent
        # Only the Cloud Type can be used as a proxy so far.
        # Adam Dybbroe, 2012-08-31
        nodata_mask = False #np.ma.masked_equal(np.ones(self.shape), 0).mask
        for key in self._projectables:
            projectable = getattr(self,  key)
            if key in ['cloudtype']:
                nodata_array = np.ma.array(projectable.data)
                nodata_mask =  np.ma.masked_equal(nodata_array, 0).mask
                break

        try:
            from pyresample import geometry
        except ImportError:
            return

        tiepoint_grid = False
        if hasattr(self, "row_indices") and hasattr(self, "column_indices"):
            column_indices = self.column_indices.data
            row_indices = self.row_indices.data
            tiepoint_grid = True

        interpolate = False
        if hasattr(self, "lon") and hasattr(self, "lat"):
            if 'intercept' in self.lon.info:
                offset_lon = self.lon.info["intercept"]
            elif 'offset' in self.lon.info:
                offset_lon = self.lon.info["offset"]
            if 'gain' in self.lon.info:
                gain_lon = self.lon.info["gain"]
            lons = self.lon.data * gain_lon + offset_lon

            if 'intercept' in self.lat.info:
                offset_lat = self.lat.info["intercept"]
            elif 'offset' in self.lat.info:
                offset_lat = self.lat.info["offset"]
            if 'gain' in self.lat.info:
                gain_lat = self.lat.info["gain"]
            lats = self.lat.data * gain_lat + offset_lat

            if lons.shape != self.shape or lats.shape != self.shape:
                # Data on tiepoint grid:
                interpolate = True
                if not tiepoint_grid:
                    errmsg = ("Interpolation needed but insufficient" + 
                              "information on the tiepoint grid")
                    raise IOError(errmsg)
            else:
                # Geolocation available on the full grid:
                # We neeed to mask out nodata (VIIRS Bow-tie deletion...)
                # We do it for all instruments, checking only against the nodata
                lons = np.ma.masked_array(lons, nodata_mask)
                lats = np.ma.masked_array(lats, nodata_mask)

                self.area = geometry.SwathDefinition(lons=lons, lats=lats)


        elif hasattr(self, "region") and self.region.data["area_extent"].any():
            region = self.region.data
            proj_dict = dict([elt.split('=')
                              for elt in region["pcs_def"].split(',')])
            self.area = geometry.AreaDefinition(region["id"],
                                                region["name"],
                                                region["proj_id"],
                                                proj_dict,
                                                region["xsize"],
                                                region["ysize"],
                                                region["area_extent"])

        if interpolate:
            from geotiepoints import SatelliteInterpolator
        
            cols_full = np.arange(self.shape[1])
            rows_full = np.arange(self.shape[0])

            satint = SatelliteInterpolator((lons, lats),
                                           (row_indices, 
                                            column_indices),
                                           (rows_full, cols_full))
            #satint.fill_borders("y", "x")
            lons, lats = satint.interpolate()

            self.area = geometry.SwathDefinition(lons=lons, lats=lats)



    def project(self, coverage):
        """Project what can be projected in the product.
        """

        import copy
        res = copy.copy(self)

        # Project the data
        for var in self._projectables:
            LOG.info("Projecting " + str(var))
            res.__dict__[var] = copy.copy(self.__dict__[var])
            res.__dict__[var].data = coverage.project_array(
                self.__dict__[var].data)

        # Take care of geolocation

        res.region = copy.copy(self.region)

        region = copy.copy(res.region.data)
        area = coverage.out_area
        try:
            # It's an area
            region["area_extent"] = np.array(area.area_extent)
            region["xsize"] = area.x_size
            region["ysize"] = area.y_size
            region["xscale"] = area.pixel_size_x
            region["yscale"] = area.pixel_size_y
            region["lon_0"] = area.proj_dict.get("lon_0", 0)
            region["lat_0"] = area.proj_dict.get("lat_0", 0)
            region["lat_ts"] = area.proj_dict.get("lat_ts", 0)
            region["name"] = area.name
            region["id"] = area.area_id
            region["pcs_id"] = area.proj_id
            pcs_def = ",".join([key + "=" + val
                                for key, val in area.proj_dict.iteritems()])
            region["pcs_def"] = pcs_def
            res.region.data = region

            # If switching to area representation, try removing lon and lat
            try:
                delattr(res, "lon")
                res._keys.remove("lon")
                delattr(res, "lat")
                res._keys.remove("lat")
            except AttributeError:
                pass
            
        except AttributeError:
            # It's a swath
            lons, scale_factor, add_offset, no_data = \
                  pack_signed(area.lons[:], np.int16)
            res.lon = InfoObject()
            res.lon.data = lons
            res.lon.info["description"] = "geographic longitude (deg)"
            res.lon.info["intercept"] = add_offset
            res.lon.info["gain"] = scale_factor
            res.lon.info["no_data_value"] = no_data
            if "lon" not in res._keys:
                res._keys.append("lon")

            lats, scale_factor, add_offset, no_data = \
                  pack_signed(area.lats[:], np.int16)
            res.lat = InfoObject()
            res.lat.data = lats
            res.lat.info["description"] = "geographic latitude (deg)"
            res.lat.info["intercept"] = add_offset
            res.lat.info["gain"] = scale_factor
            res.lat.info["no_data_value"] = no_data
            if "lat" not in res._keys:
                res._keys.append("lat")
            # Remove region parameters if switching from area
            region["area_extent"] = np.zeros(4)
            region["xsize"] = 0
            region["ysize"] = 0
            region["xscale"] = 0
            region["yscale"] = 0
            region["lon_0"] = 0
            region["lat_0"] = 0
            region["lat_ts"] = 0
            region["name"] = ""
            region["id"] = ""
            region["pcs_id"] = ""
            region["pcs_def"] = ""
            res.region.data = region
        return res

    def write(self, filename):
        """Write product in hdf format to *filename*
        """
        
        LOG.debug("Writing to " + filename)
        h5f = h5py.File(filename, "w")

        for dataset in self._projectables:
            dset = h5f.create_dataset(dataset, data=getattr(self, dataset).data,
                                      compression='gzip', compression_opts=6)
            for key, value in getattr(self, dataset).info.iteritems():
                dset.attrs[key] = value

        for thing in self._keys:
            try:
                dset = h5f.create_dataset(thing, data=getattr(self, thing).data,
                                          compression='gzip', compression_opts=6)
                for key, value in getattr(self, thing).info.iteritems():
                    dset.attrs[key] = value
            except AttributeError:
                h5f[thing] = getattr(self, thing)

        for key, value in self._md.iteritems():
            if key in ["time_slot", "satellite"]:
                continue
            h5f.attrs[key] = value

        for (key, skey), value in self._refs.iteritems():
            h5f[key].attrs[skey] = h5f[value].ref

        h5f.close()

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return len(self._projectables) > 0

class CloudType(NwcSafPpsChannel):

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CloudType"

class CloudTopTemperatureHeight(NwcSafPpsChannel):

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CTTH"

class CloudMask(NwcSafPpsChannel):

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CMa"

class PrecipitationClouds(NwcSafPpsChannel):

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "PC"

class CloudPhysicalProperties(NwcSafPpsChannel):

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CPP"



def load(scene, geofilename=None, **kwargs):
    del kwargs

    import glob

    lonlat_is_loaded = False

    products = []
    if "CTTH" in scene.channels_to_load:
        products.append("CTTH")
    if "CloudType" in scene.channels_to_load:
        products.append("CT")
    if "CMa" in scene.channels_to_load:
        products.append("CMA")
    if "PC" in scene.channels_to_load:
        products.append("PC")
    if "CPP" in scene.channels_to_load:
        products.append("CPP")

    if len(products) == 0:
        return


    try:
        area_name = scene.area_id or scene.area.area_id
    except AttributeError:
        area_name = "satproj_?????_?????"


    conf = ConfigParser.ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, scene.fullname+".cfg"))
    directory = conf.get(scene.instrument_name+"-level3", "dir")
    try:
        geodir = conf.get(scene.instrument_name+"-level3", "geodir")
    except NoOptionError:
        LOG.warning("No option 'geodir' in level3 section")
        geodir = None

    filename = conf.get(scene.instrument_name+"-level3", "filename",
                        raw=True)
    pathname_tmpl = os.path.join(directory, filename)

    if not geofilename and geodir:
        # Load geo file from config file:
        try:
            if not scene.orbit:
                orbit = ""
            else:
                orbit = scene.orbit
            geoname_tmpl = conf.get(scene.instrument_name+"-level3", 
                                    "geofilename", raw=True)
            filename_tmpl = (scene.time_slot.strftime(geoname_tmpl)
                             %{"orbit": orbit.zfill(5) or "*",
                               "area": area_name,
                               "satellite": scene.satname + scene.number})

            file_list = glob.glob(os.path.join(geodir, filename_tmpl))
            if len(file_list) > 1:
                LOG.warning("More than 1 file matching for geoloaction: "
                            + str(file_list))
            elif len(file_list) == 0:
                LOG.warning("No geolocation file matching!: " + filename_tmpl)
            else:
                geofilename = file_list[0]
        except NoOptionError:
            geofilename = None


    classes = {"CTTH": CloudTopTemperatureHeight,
               "CT": CloudType,
               "CMA": CloudMask,
               "PC": PrecipitationClouds,
               "CPP": CloudPhysicalProperties
               }

    nodata_mask = False

    chn = None
    for product in products:
        LOG.debug("Loading " + product)
        if not scene.orbit:
            orbit = ""
        else:
            orbit = scene.orbit
        filename_tmpl = (scene.time_slot.strftime(pathname_tmpl)
                         %{"orbit": orbit.zfill(5) or "*",
                           "area": area_name,
                           "satellite": scene.satname + scene.number,
                           "product": product})
    
        file_list = glob.glob(filename_tmpl)
        if len(file_list) > 1:
            LOG.warning("More than 1 file matching for " + product + "! "
                        + str(file_list))
            continue
        elif len(file_list) == 0:
            LOG.warning("No " + product + " matching!: " + filename_tmpl)
            continue
        else:
            filename = file_list[0]

            chn = classes[product]()
            chn.read(filename, lonlat_is_loaded==False)
            scene.channels.append(chn)


        # Setup geolocation
        # We need a no-data mask from one of the projectables to
        # mask out bow-tie deletion pixels from the geolocation array
        # So far only relevant for VIIRS.
        # Preferably the lon-lat data in the PPS VIIRS geolocation
        # file should already be masked. 
        # The no-data values in the products are not only where geo-location is absent
        # Only the Cloud Type can be used as a proxy so far.
        # Adam Dybbroe, 2012-08-31
        if hasattr(chn, '_projectables'):
            for key in chn._projectables:
                projectable = getattr(chn,  key)
                if key in ['ct']:
                    nodata_array = np.ma.array(projectable.data)
                    nodata_mask =  np.ma.masked_equal(\
                        nodata_array, projectable.info["_FillValue"]).mask
                    break
        else:
            LOG.warning("Channel has no '_projectables' member." + 
                        " No nodata-mask set...")

    if chn is None:
        return

    # Is this safe!? AD 2012-08-25
    shape = chn.shape

    interpolate = False
    if geofilename:
        geodict = get_lonlat(geofilename)
        lons, lats = geodict['lon'], geodict['lat']
        if lons.shape != shape or lats.shape != shape:
            interpolate = True
            row_indices = geodict['row_indices']
            column_indices = geodict['column_indices']

        lonlat_is_loaded = True
    else:
        LOG.warning("No Geo file specified: " + 
                    "Geolocation will be loaded from product")


    if lonlat_is_loaded:
        if interpolate:
            from geotiepoints import SatelliteInterpolator
        
            cols_full = np.arange(shape[1])
            rows_full = np.arange(shape[0])

            satint = SatelliteInterpolator((lons, lats),
                                           (row_indices, 
                                            column_indices),
                                           (rows_full, cols_full))
            #satint.fill_borders("y", "x")
            lons, lats = satint.interpolate()

        try:
            from pyresample import geometry
            lons = np.ma.masked_array(lons, nodata_mask)
            lats = np.ma.masked_array(lats, nodata_mask)
            scene.area = geometry.SwathDefinition(lons=lons, 
                                                  lats=lats)
        except ImportError:
            scene.area = None
            scene.lat = lats
            scene.lon = lons

            
    LOG.info("Loading PPS parameters done.")


def get_lonlat(filename):
    """Read lon,lat from hdf5 file"""
    import h5py
    LOG.debug("Geo File = " + filename)

    h5f = h5py.File(filename, 'r')

    # We neeed to mask out nodata (VIIRS Bow-tie deletion...)
    # We do it for all instruments, checking only against the nodata
    nodata = h5f['where']['lon']['what'].attrs['nodata']
    gain = h5f['where']['lon']['what'].attrs['gain']
    offset = h5f['where']['lon']['what'].attrs['offset']

    longitudes = np.ma.array(h5f['where']['lon']['data'].value)
    lons = np.ma.masked_equal(longitudes, nodata) * gain + offset
    latitudes = np.ma.array(h5f['where']['lat']['data'].value)
    lats = np.ma.masked_equal(latitudes, nodata) * gain + offset

    col_indices = None
    row_indices = None
    if "column_indices" in h5f["where"].keys():
        col_indices = h5f['/where/column_indices'].value
    if "row_indices" in h5f["where"].keys():
        row_indices = h5f['/where/row_indices'].value

    h5f.close()
    return {'lon': lons, 
            'lat': lats, 
            'col_indices': col_indices, 'row_indices':row_indices}
    #return lons, lats



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2016.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Setup file for mpop.
"""
import os.path
from setuptools import setup
import imp

version = imp.load_source('mpop.version', 'mpop/version.py')

BASE_PATH = os.path.sep.join(os.path.dirname(
    os.path.realpath(__file__)).split(os.path.sep))

requires = ['numpy >=1.4.1', 'pyresample']

try:
    from PIL import Image
except ImportError:
    requires.append("pillow")

NAME = 'mpop'

setup(name=NAME,
      version=version.__version__,
      description='Meteorological post processing package',
      author='Martin Raspaud',
      author_email='martin.raspaud@smhi.se',
      classifiers=["Development Status :: 5 - Production/Stable",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: GNU General Public License v3 " +
                   "or later (GPLv3+)",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Topic :: Scientific/Engineering"],
      url="https://github.com/mraspaud/mpop",
      test_suite='mpop.tests.suite',
      packages=['mpop', 'mpop.satellites', 'mpop.instruments', 'mpop.satin',
                'mpop.satout', 'mpop.saturn', 'mpop.imageo',
                'mpop.imageo.formats'],
      data_files=[(os.path.join('etc', 'pytroll'),
                   [os.path.join('etc', 'geo_image.cfg'),
                    os.path.join('etc', 'eps_avhrrl1b_6.5.xml')])],
      zip_safe=False,
      install_requires=requires,
      tests_require=["mock", 'pyorbital >= v0.2.3'],
      extras_require={'xRIT': ['mipp >= 0.6.0'],
                      'hdf_eos': ['pyhdf'],
                      'viirs': ['h5py'],
                      'nc': ['netCDF4'],
                      'proj': ['pyresample'],
                      'pyspectral': ['pyspectral'],
                      'pyorbital': ['pyorbital >= v0.2.3'],
                      'gac_lac': ['pygac']}

      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015, 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Panu Lahtinen <panu.lahtinen@fmi.fi>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""PPS netcdf cloud product reader
"""
import os.path
from ConfigParser import ConfigParser
from ConfigParser import NoOptionError
from glob import glob
from datetime import datetime
import numpy as np
import h5py

from trollsift import Parser
import mpop.channel
from mpop import CONFIG_PATH
from mpop.plugin_base import Reader

import logging
LOG = logging.getLogger(__name__)


class InconsistentDataDimensions(Exception):

    """Exception for inconsistent dimensions in the data"""
    pass


def unzip_file(filename):
    """Unzip the file if file is bzipped = ending with 'bz2'"""

    import tempfile
    import bz2
    if filename.endswith('bz2'):
        bz2file = bz2.BZ2File(filename)
        tmpfilename = tempfile.mktemp()
        try:
            ofpt = open(tmpfilename, 'wb')
            ofpt.write(bz2file.read())
            ofpt.close()
        except IOError:
            import traceback
            traceback.print_exc()
            LOG.info("Failed to read bzipped file %s", str(filename))
            os.remove(tmpfilename)
            return None

        return tmpfilename

    return None


class GeolocationFlyweight(object):

    """Flyweight-thingy for geolocation:
    http://yloiseau.net/articles/DesignPatterns/flyweight/
    """

    def __init__(self, cls):
        self._cls = cls
        self._instances = dict()

    def __call__(self, *args, **kargs):
        """
        we assume that this is only used for the gelocation object,
        filenames are listed in the second argument

        """
        return self._instances.setdefault(tuple(args[1]),
                                          self._cls(*args, **kargs))

    def clear_cache(self):
        """Clear cache"""
        del self._instances
        self._instances = dict()


#@GeolocationFlyweight
class PpsGeolocationData(object):

    '''Class for holding PPS geolocation data'''

    def __init__(self, shape, granule_lengths, filenames):
        self.filenames = filenames
        self.shape = shape
        self.granule_lengths = granule_lengths
        self.longitudes = None
        self.latitudes = None
        self.row_indices = None
        self.col_indices = None
        self.mask = None

    def read(self):
        """
        Read longitudes and latitudes from geo filenames and assemble
        """

        if self.longitudes is not None:
            return self

        self.longitudes = np.empty(self.shape,
                                   dtype=np.float32)
        self.latitudes = np.empty(self.shape,
                                  dtype=np.float32)
        self.mask = np.zeros(self.shape,
                             dtype=np.bool)

        swath_index = 0

        for idx, filename in enumerate(self.filenames):

            y0_ = swath_index
            y1_ = swath_index + self.granule_lengths[idx]
            swath_index = swath_index + self.granule_lengths[idx]

            get_lonlat_into(filename,
                            self.longitudes[y0_:y1_, :],
                            self.latitudes[y0_:y1_, :],
                            self.mask[y0_:y1_, :])

        self.longitudes = np.ma.array(self.longitudes,
                                      mask=self.mask,
                                      copy=False)
        self.latitudes = np.ma.array(self.latitudes,
                                     mask=self.mask,
                                     copy=False)

        LOG.debug("Geolocation read in for %s", str(self))
        return self


class HDF5MetaData(object):

    """

    Small class for inspecting a HDF5 file and retrieve its metadata/header
    data. It is developed for JPSS/NPP data but is really generic and should
    work on most other hdf5 files.

    Supports

    """

    def __init__(self, filename):
        self.metadata = {}
        self.filename = filename
        if not os.path.exists(filename):
            raise IOError("File %s does not exist!" % filename)

    def read(self):
        """Read the metadata"""
        filename = self.filename
        unzipped = unzip_file(filename)
        if unzipped:
            filename = unzipped

        with h5py.File(filename, 'r') as h5f:
            h5f.visititems(self.collect_metadata)
            self._collect_attrs('/', h5f.attrs)

        if unzipped:
            os.remove(unzipped)

        return self

    def _collect_attrs(self, name, attrs):
        """Collect atributes"""
        for key in attrs.keys():
            # Throws a TypeError if key==DIMENSION_LIST and the value
            # is accessed
            # Observed at FMI (Panu) - maybe hdf5 version specific?
            # Should preferably be handled elsewhere and not in this generic class
            # FIXME!
            if key in ['DIMENSION_LIST']:
                continue
            value = np.squeeze(attrs[key])
            if issubclass(value.dtype.type, str):
                self.metadata["%s/attr/%s" % (name, key)] = str(value)
            else:
                self.metadata["%s/attr/%s" % (name, key)] = value

    def collect_metadata(self, name, obj):
        """Collect metadata"""
        if isinstance(obj, h5py.Dataset):
            self.metadata["%s/shape" % name] = obj.shape
        self._collect_attrs(name, obj.attrs)

    def __getitem__(self, key):

        long_key = None
        for mkey in self.metadata.keys():
            if mkey.endswith(key):
                if long_key is not None:
                    raise KeyError("Multiple keys called %s" % key)
                long_key = mkey
                break
        return self.metadata[long_key]

    def keys(self):
        """Return metadata dictionary keys"""
        return self.metadata.keys()

    def get_data_keys(self):
        """Get data keys from the metadata"""
        data_keys = []
        for key in self.metadata.keys():
            if key.endswith("/shape"):
                data_key = key.split("/shape")[0]
                data_keys.append(data_key)
        return data_keys

    def get_data_keys_and_shapes(self):
        """Get data keys and array shapes from the metadata"""
        data_keys = {}
        for key in self.metadata.keys():
            if key.endswith("/shape"):
                data_key = key.split("/shape")[0]
                shape = self.metadata[key]
                data_keys[data_key] = shape

        return data_keys


class PPSMetaData(HDF5MetaData):

    """Class for holding PPS metadata"""

    def get_shape(self):
        """Get array shapes from metadata"""
        n_x = 0
        n_y = 0
        for key in self.metadata:
            if key.endswith('nx/shape'):
                n_x = self.metadata[key][0]
            if key.endswith('ny/shape'):
                n_y = self.metadata[key][0]

        return n_x, n_y

    def get_header_info(self):
        """Get platform name, orbit number and time slot as dictionary"""
        info = {}
        for key in self.metadata:
            if key.endswith('platform'):
                info['platform_name'] = self.metadata[key]
            elif key.endswith('orbit_number'):
                info['orbit'] = self.metadata[key]
            elif key.endswith('time_coverage_start'):
                info['time_slot'] = datetime.strptime(self.metadata[key][:-2],
                                                      "%Y%m%dT%H%M%S")
        return info

    def get_dataset_attributes(self, var_name):
        """Get dataset attributes"""
        retv = {}
        for key in self.metadata:
            if key.split('/')[0] == var_name and key.find('attr') > 0:
                dictkey = key.split('/')[-1]
                if dictkey in ['DIMENSION_LIST']:
                    continue
                retv[dictkey] = self.metadata[key]

        return retv

    def get_root_attributes(self):
        """Get attributes of the root directory"""
        retv = {}
        for key in self.metadata:
            if key.startswith('//attr'):
                dictkey = key.split('/')[-1]
                retv[dictkey] = self.metadata[key]

        return retv


def get_filenames(scene, products, conf, time_interval, area_name):
    """Get list of filenames within time interval"""

    filename = conf.get(scene.instrument_name + "-level3",
                        "cloud_product_filename",
                        raw=True,
                        vars=os.environ)
    directory = conf.get(scene.instrument_name + "-level3",
                         "cloud_product_dir",
                         vars=os.environ)
    pathname_tmpl = os.path.join(directory, filename)
    starttime, endtime = time_interval

    if not scene.orbit:
        orbit = ""
    else:
        orbit = scene.orbit

    flist_allproducts = []
    for product in products:
        values = {"area": area_name,
                  "satellite": scene.satname + scene.number,
                  "product": product}

        if endtime:
            # Okay, we need to check for more than one granule/swath!
            # First get all files with all times matching in directory:
            values["orbit"] = '?????'
            filename_tmpl = os.path.join(directory,
                                         globify_date(filename) % values)

        else:
            values["orbit"] = str(orbit).zfill(5) or "*"
            filename_tmpl = scene.time_slot.strftime(
                pathname_tmpl) % values

        LOG.debug("File path = %s", str(filename_tmpl))
        file_list = glob(filename_tmpl)
        if len(file_list) == 0:
            LOG.warning("No %s product matching", str(product))
        elif len(file_list) > 1 and not endtime:
            LOG.warning("More than 1 file matching for %s: %s",
                        str(product), str(file_list))
            file_list = []
        elif len(file_list) > 1:
            file_list = extract_filenames_in_time_window(
                file_list, starttime, endtime)

        if len(file_list) == 0:
            LOG.warning("No files found matching time window for product %s",
                        product)

        flist_allproducts = flist_allproducts + file_list

    return flist_allproducts


def extract_filenames_in_time_window(file_list, starttime, endtime):
    """Extract the filenames with time inside the time interval specified.
    NB! Only tested for EARS-NWC granules. This does not support assembling
    several locally received full swaths"""

    # New EARS-NWC filenames:
    # Ex.:
    # W_XX-EUMETSAT-Darmstadt,SING+LEV+SAT,NOAA19+CT_C_EUMS_20150819124700_\
    #  33643.nc.bz2
    pnew = Parser(EARS_PPS_FILE_MASK)

    # Old EARS-NWC filenames:
    # Ex.:
    # ctth_20130910_205300_metopb.h5.bz2
    pold = Parser("{product:s}_{starttime:%Y%m%d_%H%M}00_{platform_name:s}.h5"
                  "{compression:s}")

    plocal = Parser(LOCAL_PPS_FILE_MASK)

    valid_filenames = []
    valid_times = []
    LOG.debug("Time window: (%s, %s)", str(starttime), str(endtime))
    for fname in file_list:
        try:
            data = pnew.parse(os.path.basename(fname))
        except ValueError:
            try:
                data = pold.parse(os.path.basename(fname))
            except ValueError:
                data = plocal.parse(os.path.basename(fname))

        if (data['starttime'] >= starttime and
                data['starttime'] < endtime):
            valid_filenames.append(fname)
            valid_times.append(data['starttime'])
            LOG.debug("Start time %s inside window", str(data['starttime']))
        else:
            pass

    # Can we rely on the files being sorted according to time?
    # Sort the filenames according to time:
    vtimes = np.array(valid_times)
    idx = np.argsort(vtimes)
    vfiles = np.array(valid_filenames)
    return np.take(vfiles, idx).tolist()


class InfoObject(object):

    """Simple data and info container.
    """

    def __init__(self):
        self.info = {}
        self.data = None


class NwcSafPpsChannel(mpop.channel.GenericChannel):

    """Class for NWC-SAF PPS channel data"""

    def __init__(self):
        mpop.channel.GenericChannel.__init__(self)
        self.mda = {}
        self._projectables = []
        self.shape = None
        self.granule_lengths = None
        self.filenames = None

        self.platform_name = None
        self.begin_time = None
        self.end_time = None
        self.orbit_begin = None
        self.orbit_end = None

    def read(self, pps_product):
        """Read the PPS v2014 formated data"""

        LOG.debug("Read the PPS product data...")

        self._projectables = pps_product.projectables
        self.granule_lengths = pps_product.granule_lengths
        self.shape = pps_product.shape
        self.filenames = pps_product.filenames
        self.orbit_begin = pps_product.orbit_begin
        self.orbit_end = pps_product.orbit_end
        self.platform_name = pps_product.platform_name
        self.begin_time = pps_product.begin_time
        self.end_time = pps_product.end_time

        # Take the metadata of the first granule and store as global
        #self.mda = pps_product.metadata[0].metadata
        mda = pps_product.metadata[0]
        self.mda = mda.metadata
        self.mda.update(mda.get_root_attributes())

        for var_name in pps_product.mda.keys():
            setattr(self, var_name, InfoObject())
            # Fill the info dict...
            getattr(self, var_name).info = mda.get_dataset_attributes(var_name)
            try:
                getattr(self, var_name).data = self.mda[var_name]
            except KeyError:
                continue

        for var_name in self._projectables:
            setattr(self, var_name, InfoObject())
            # Fill the info dict...
            getattr(self, var_name).info = mda.get_dataset_attributes(var_name)
            getattr(self, var_name).data = \
                np.ma.masked_array(pps_product.raw_data[var_name],
                                   mask=pps_product.mask[var_name],
                                   fill_value=pps_product.fill_value[var_name])

        return

    def project(self, coverage):
        """Project the data"""
        LOG.debug("Projecting channel %s...", self.name)
        import copy
        res = copy.copy(self)

        # Project the data
        for var in self._projectables:
            LOG.info("Projecting %s", str(var))
            res.__dict__[var] = copy.copy(self.__dict__[var])
            res.__dict__[var].data = coverage.project_array(
                self.__dict__[var].data)

        res.name = self.name
        res.resolution = self.resolution
        res.filled = True
        res.area = coverage.out_area
        return res

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return len(self._projectables) > 0

    def save(self, filename, old=True, **kwargs):
        """Save to old format"""
        del kwargs
        if old:
            try:
                from nwcsaf_formats.ppsv2014_to_oldformat import write_product
                write_product(self, filename)
            except ImportError:
                LOG.error("Could not save to old format")
                raise
        else:
            raise NotImplementedError("Can't save to new pps format yet.")


class PPSProductData(object):

    """Placeholder for the PPS netCDF product data. Reads the
    netCDF files using h5py. One file for each product and granule/swath.

    """

    def __init__(self, filenames):

        self.filenames = filenames
        self.mda = {}
        self.projectables = []
        self._keys = []
        self._refs = {}
        self.shape = None
        self.product_name = 'unknown'
        self.platform_name = None

        self.begin_time = None
        self.end_time = None
        self.orbit_begin = None
        self.orbit_end = None

        self.metadata = []

        self.raw_data = {}
        self.mask = {}
        self.fill_value = {}

        self.granule_lengths = []

    def read(self):
        """Read the PPS v2014 formatet data"""

        self._read_metadata()

        for key in self.raw_data:
            LOG.debug("Shape of data: %s", str(self.raw_data[key].shape))
            break

        self._read_data()

        return self

    def _set_members(self, hdd):
        '''Set platform_name, time_slot and orbit class members'''
        if not self.platform_name and 'platform_name' in hdd:
            self.platform_name = hdd['platform_name']
        if not self.begin_time and 'time_slot' in hdd:
            self.begin_time = hdd['time_slot']
        if 'time_slot' in hdd:
            self.end_time = hdd['time_slot']
        if not self.orbit_begin and 'orbit' in hdd:
            self.orbit_begin = int(hdd['orbit'])
        if 'orbit' in hdd:
            self.orbit_end = int(hdd['orbit'])

    def _read_metadata(self):
        """Read metadata from all the files"""
        LOG.debug("Filenames: %s", str(self.filenames))

        swath_length = 0
        swath_width = None
        for fname in self.filenames:
            LOG.debug("Get and append metadata from file: %s", str(fname))
            mda = PPSMetaData(fname).read()
            # Set the product_name variable:
            try:
                self.product_name = mda['product_name']
            except KeyError:
                LOG.warning("No product_name in file!")

            width, granule_length = mda.get_shape()
            hdd = mda.get_header_info()
            self._set_members(hdd)

            self.metadata.append(mda)
            self.granule_lengths.append(granule_length)

            if swath_width:
                if swath_width != width:
                    raise InconsistentDataDimensions('swath_width not the same '
                                                     'between granules: %d %d',
                                                     swath_width, width)

            swath_width = width
            swath_length = swath_length + granule_length

        # Take the first granule, and find what data fields it contains
        # and assume all granules have those same data fields:
        mda = self.metadata[0]
        dks = mda.get_data_keys_and_shapes()
        geolocation_fields = ['lon', 'lat', 'lat_reduced', 'lon_reduced']
        coordinate_fields = ['nx', 'nx_reduced', 'ny', 'ny_reduced']

        for key in dks:
            if key in geolocation_fields + coordinate_fields:
                LOG.debug("Key = %s", str(key))
                continue
            shape = dks[key]
            if len(shape) == 3 and shape[0] == 1:
                shape = shape[1], shape[2]
            if shape == (self.granule_lengths[0], swath_width):
                self.projectables.append(key)
            else:
                self.mda.update({key: dks[key]})

            # LOG.debug("Key, shape, granule_length, swath_width: %s %s %s %s",
            #          str(key), str(shape),
            #          str(self.granule_lengths[0]), str(swath_width))

        # initiate data arrays
        self.shape = swath_length, swath_width

        # for field in dks:
        #     if field in geolocation_fields + coordinate_fields:
        #         continue
        # try:
        #     dtype = mda[field + '/attr/valid_range'].dtype
        #     self.raw_data[str(field)] = np.zeros(self.shape, dtype=dtype)
        #     self.mask[field] = np.zeros(self.shape, dtype=np.bool)
        # except KeyError:
        #     continue
        for field in self.projectables:
            dtype = mda[field + '/attr/valid_range'].dtype
            try:
                if not (np.equal(1.0 + mda[field + '/attr/add_offset'], 1.0) and
                        np.equal(1.0 * mda[field + '/attr/scale_factor'], 1.0)):
                    dtype = np.float32
            except KeyError:
                pass
            self.raw_data[str(field)] = np.zeros(self.shape, dtype=dtype)
            self.mask[field] = np.zeros(self.shape, dtype=np.bool)

    def _read_data(self):
        """Loop over all granules and read one granule at a time and
        fill the data arrays"""

        LOG.debug("Read all %s product files...", self.product_name)
        swath_index = 0
        for idx, mda in enumerate(self.metadata):
            del mda
            filename = self.filenames[idx]

            unzipped = unzip_file(filename)
            if unzipped:
                filename = unzipped

            h5f = h5py.File(filename, 'r')

            variables = {}
            for key, item in h5f.items():
                if item.attrs.get("CLASS") != 'DIMENSION_SCALE':
                    variables[key] = item

            # processed variables
            processed = set()
            non_processed = set(variables.keys()) - processed

            fields = {}
            for var_name in non_processed:
                if var_name in ['lon', 'lat', 'lon_reduced', 'lat_reduced']:
                    continue

                var = variables[var_name]
                if ("standard_name" not in var.attrs.keys() and
                        "long_name" not in var.attrs.keys()):
                    LOG.warning("Data field %s is lacking both "
                                "standard_name and long_name",
                                var_name)
                    continue

                if var_name not in self.projectables:
                    self.metadata[idx].metadata[var_name] = var[:]
                    continue

                data = var[:]
                if len(data.shape) == 3 and data.shape[0] == 1:
                    data = data[0]

                if 'valid_range' in var.attrs.keys():
                    data = np.ma.masked_outside(
                        data, *var.attrs['valid_range'])
                elif '_FillValue' in var.attrs.keys():
                    data = np.ma.masked_where(data, var.attrs['_FillValue'])
                if "scale_factor" in var.attrs.keys() and \
                   "add_offset" in var.attrs.keys():
                    dataset = (data * var.attrs.get("scale_factor", 1)
                               + var.attrs.get("add_offset", 0))
                else:
                    dataset = data.copy()
                    if '_FillValue' in var.attrs.keys():
                        dataset.fill_value = var.attrs['_FillValue'][0]

                fields[var_name] = dataset

                LOG.debug("long_name: %s", str(var.attrs['long_name']))
                LOG.debug("Var = %s, shape = %s",
                          str(var_name), str(dataset.shape))

                processed |= set([var_name])

            non_processed = set(variables.keys()) - processed
            if len(non_processed) > 0:
                LOG.warning("Remaining non-processed variables: %s",
                            str(non_processed))

            h5f.close()
            if unzipped:
                os.remove(unzipped)

            y0_ = swath_index
            y1_ = swath_index + self.granule_lengths[idx]
            swath_index = swath_index + self.granule_lengths[idx]

            for key in self.raw_data.keys():
                if key not in self.projectables:
                    continue
                try:
                    self.raw_data[key][y0_:y1_, :] = fields[key].data
                    self.mask[key][y0_:y1_, :] = fields[key].mask
                    self.fill_value[key] = fields[key].fill_value
                except ValueError:
                    LOG.exception('Mismatch in dimensions: y0_, y1_, '
                                  'fields[key].data.shape: %s %s %s',
                                  str(y0_), str(y1_),
                                  str(fields[key].data.shape))
                    raise

        return

GEO_PRODUCT_NAME_DEFAULT = 'CMA'
PPS_PRODUCTS = set(['CMA', 'CT', 'CTTH', 'PC', 'CPP'])
LOCAL_PPS_FILE_MASK = ('S_NWC_{product:s}_{platform_name:s}_{orbit:5d}_' +
                       '{starttime:%Y%m%dT%H%M%S}{dummy:1d}Z_' +
                       '{starttime:%Y%m%dT%H%M%S}{dummy2:1d}Z.nc{compression:s}')
EARS_PPS_FILE_MASK = ("W_XX-EUMETSAT-Darmstadt,SING+LEV+SAT,{platform_name:s}+"
                      "{product:s}_C_EUMS_{starttime:%Y%m%d%H%M}00_{orbit:05d}.nc"
                      "{compression:s}")
# Old EARS-NWC filenames:
# Ex.:
# ctth_20130910_205300_metopb.h5.bz2
EARS_OLD_PPS_FILE_MASK = ("{product:s}_{starttime:%Y%m%d_%H%M}00_" +
                          "{platform_name:s}.h5{compression:s}")


class PPSReader(Reader):

    """Reader class for PPS files"""
    pformat = "nc_pps_l2"

    def __init__(self, *args, **kwargs):
        Reader.__init__(self, *args, **kwargs)
        # Source of the data, 'local' or 'ears'
        self._source = None
        # Parser for getting info from the file names
        self._parser = None
        # Satellite config
        self._config = None
        # Location of geolocation files, required for 'local' products
        self._cloud_product_geodir = None
        # Name of the product having geolocation for 'local' products
        self._geolocation_product_name = None

    def _read_config(self, sat_name, instrument_name):
        '''Read config for the satellite'''

        if self._config:
            return

        self._config = ConfigParser()
        configfile = os.path.join(CONFIG_PATH, sat_name + ".cfg")
        LOG.debug("Read configfile %s", configfile)
        self._config.read(configfile)

        try:
            self._cloud_product_geodir = \
                self._config.get(instrument_name + "-level3",
                                 "cloud_product_geodir",
                                 raw=True,
                                 vars=os.environ)
        except NoOptionError:
            pass

        LOG.debug("cloud_product_geodir = %s", self._cloud_product_geodir)

        try:
            self._geolocation_product_name = \
                self._config.get(instrument_name + "-level3",
                                 "geolocation_product_name",
                                 raw=True,
                                 vars=os.environ)
        except NoOptionError:
            if self._source != 'ears':
                LOG.warning("No geolocation product name given in config, "
                            "using default: %s", GEO_PRODUCT_NAME_DEFAULT)
                self._geolocation_product_name = GEO_PRODUCT_NAME_DEFAULT

    def _determine_prod_and_geo_files(self, prodfilenames):
        """From the list of product files and the products to load determine the
        product files and the geolocation files that will be considered when
        reading the data

        """

        # geofiles4product is a dict listing all geo-locations files applicable
        # for each product.
        # prodfiles4product is a dict listing all product files for a given
        # product name

        prodfiles4product = {}
        geofiles4product = {}
        if prodfilenames:
            if not isinstance(prodfilenames, (list, set, tuple)):
                prodfilenames = [prodfilenames]
            for fname in prodfilenames:
                # Only standard NWCSAF/PPS and EARS-NWC naming accepted!
                # No support for old file names (< PPSv2014)
                if (os.path.basename(fname).startswith("S_NWC") or
                        os.path.basename(fname).startswith("W_XX-EUMETSAT")):
                    if not self._parser:
                        if os.path.basename(fname).startswith("S_NWC"):
                            self._source = 'local'
                            self._parser = Parser(LOCAL_PPS_FILE_MASK)
                        else:
                            self._source = 'ears'
                            self._parser = Parser(EARS_PPS_FILE_MASK)
                else:
                    LOG.info("Unrecognized NWCSAF/PPS file: %s", fname)
                    continue

                parse_info = self._parser.parse(os.path.basename(fname))
                prodname = parse_info['product']

                if prodname not in prodfiles4product:
                    prodfiles4product[prodname] = []

                prodfiles4product[prodname].append(fname)

            # Assemble geolocation information
            if self._source == 'ears':
                # For EARS data, the files have geolocation in themselves
                for prodname, fnames in prodfiles4product.iteritems():
                    geofiles4product[prodname] = fnames
            else:
                # For locally processed data, use the geolocation from
                # the product defined in config

                if self._geolocation_product_name in prodfiles4product:
                    for prodname in prodfiles4product.keys():
                        geofiles4product[prodname] = \
                            prodfiles4product[self._geolocation_product_name]
                else:
                    # If the product files with geolocation are not used,
                    # assume that they are still available on the disk.

                    if self._cloud_product_geodir is None:
                        LOG.warning("Config option 'cloud_product_geodir' is not "
                                    "available! Assuming same directory as "
                                    "products.")
                    for prodname in prodfiles4product.keys():
                        geofiles4product[prodname] = []
                        for fname in prodfiles4product[prodname]:
                            directory = self._cloud_product_geodir or \
                                os.path.abspath(fname)
                            parse_info = \
                                self._parser.parse(os.path.basename(fname))
                            fname = fname.replace(parse_info['product'],
                                                  self._geolocation_product_name)
                            fname = os.path.join(directory, fname)
                            geofiles4product[prodname].append(fname)

            # Check that each product file has a corresponding geolocation
            # file:
            '''
            if self._geolocation_product_name:
                for prod in products:
                    if prod not in geofiles4product:
                        LOG.error("No product name %s in dict "
                                  "geofiles4product!",
                                  prod)
                        continue
                    if prod not in prodfiles4product:
                        LOG.error("No product name %s in dict "
                                  "prodfiles4product!",
                                  prod)
                        continue
                    if len(geofiles4product[prod]) != \
                       len(prodfiles4product[prod]):
                        LOG.error("Mismatch in number of product files and "
                                  "matching geolocation files!")
            '''

        return prodfiles4product, geofiles4product

    def load(self, satscene, **kwargs):
        """Read data from file and load it into *satscene*.
        """

        prodfilenames = kwargs.get('filename')
        time_interval = kwargs.get('time_interval')
        if prodfilenames and time_interval:
            LOG.warning("You have specified both a list of files " +
                        "and a time interval")
            LOG.warning("Specifying a time interval will only take effect " +
                        "if no files are specified")
            time_interval = None

        products = satscene.channels_to_load & set(PPS_PRODUCTS)
        if len(products) == 0:
            LOG.debug("No PPS cloud products to load, abort")
            return

        self._read_config(satscene.fullname, satscene.instrument_name)

        LOG.info("Products to load: %s", str(products))

        # If a list of files are provided to the load call, we disregard the
        # direcorty and filename specifications/definitions in the config file.

        if not prodfilenames:
            try:
                area_name = satscene.area_id or satscene.area.area_id
            except AttributeError:
                area_name = "satproj_?????_?????"

            # Make the list of files for the requested products:
            if isinstance(time_interval, (tuple, set, list)) and \
               len(time_interval) == 2:
                time_start, time_end = time_interval
            else:
                time_start, time_end = satscene.time_slot, None

            LOG.debug(
                "Start and end times: %s %s", str(time_start), str(time_end))
            prodfilenames = get_filenames(satscene, products, self._config,
                                          (time_start, time_end), area_name)

        LOG.debug("Product files: %s", str(prodfilenames))

        retv = self._determine_prod_and_geo_files(prodfilenames)
        prodfiles4product, geofiles4product = retv

        # Reading the products
        classes = {"CTTH": CloudTopTemperatureHeight,
                   "CT": CloudType,
                   "CMA": CloudMask,
                   "PC": PrecipitationClouds,
                   "CPP": CloudPhysicalProperties
                   }
        nodata_mask = False

        read_external_geo = {}
        for product in products:
            LOG.debug("Loading %s", product)

            if product not in prodfiles4product:
                LOG.warning("No files found for product: %s", product)
                continue

            pps_band = PPSProductData(prodfiles4product[product]).read()
            chn = classes[product]()
            chn.read(pps_band)

            if not chn.name in satscene:
                LOG.info("Adding new channel %s", chn.name)
                satscene.channels.append(chn)

            # Check if geolocation is loaded:
            if not chn.area:
                read_external_geo[product] = satscene.channels[-1].name

        # Check if some 'channel'/product needs geolocation. If some
        # product does not have geolocation, get it from the
        # geofilename:
        from pyresample import geometry

        # Load geolocation
        for chn_name in read_external_geo.values():
            LOG.debug("ch_name = %s", str(chn_name))
            chn = satscene[chn_name]
            geofilenames = geofiles4product[chn_name]
            LOG.debug("Geo-files = %s", str(geofilenames))
            geoloc = PpsGeolocationData(chn.shape,
                                        chn.granule_lengths,
                                        geofilenames).read()

            try:
                satscene[chn.name].area = geometry.SwathDefinition(
                    lons=geoloc.longitudes, lats=geoloc.latitudes)

                area_name = ("swath_" + satscene.fullname + "_" +
                             str(satscene.time_slot) + "_"
                             + str(chn.shape) + "_" +
                             chn.name)
                satscene[chn.name].area.area_id = area_name
                satscene[chn.name].area_id = area_name
            except ValueError:
                LOG.exception('Failed making a SwathDefinition: ' +
                              'min,max lons,lats = (%f %f") (%f,%f)',
                              geoloc.longitudes.data.min(),
                              geoloc.longitudes.data.max(),
                              geoloc.latitudes.data.min(),
                              geoloc.latitudes.data.max())
                LOG.warning("No geolocation loaded for %s", str(chn_name))

        # PpsGeolocationData.clear_cache()

        return


class CloudType(NwcSafPpsChannel):

    """CloudType PPS channel object"""

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CT"


class CloudTopTemperatureHeight(NwcSafPpsChannel):

    """Cloud top temperature and height PPS channel object"""

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CTTH"


class CloudMask(NwcSafPpsChannel):

    """Cloud mask PPS channel object"""

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CMA"


class PrecipitationClouds(NwcSafPpsChannel):

    """Precipitation clouds PPS channel object"""

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "PC"


class CloudPhysicalProperties(NwcSafPpsChannel):

    """Cloud physical proeperties PPS channel"""

    def __init__(self):
        NwcSafPpsChannel.__init__(self)
        self.name = "CPP"


def get_lonlat_into(filename, out_lons, out_lats, out_mask):
    """Read lon,lat from hdf5 file"""
    LOG.debug("Geo File = %s", filename)

    shape = out_lons.shape

    unzipped = unzip_file(filename)
    if unzipped:
        filename = unzipped

    mda = HDF5MetaData(filename).read()

    reduced_grid = False
    h5f = h5py.File(filename, 'r')

    if "column_indices" in h5f.keys():
        col_indices = h5f["column_indices"][:]
    if "row_indices" in h5f.keys():
        row_indices = h5f["row_indices"][:]
    if "nx_reduced" in h5f:
        col_indices = h5f["nx_reduced"][:]
    if "ny_reduced" in h5f:
        row_indices = h5f["ny_reduced"][:]

    for key in mda.get_data_keys():
        if ((key.endswith("lat") or key.endswith("lon")) or
                (key.endswith("lat_reduced") or key.endswith("lon_reduced"))):
            lonlat = h5f[key]
            fillvalue = lonlat.attrs["_FillValue"]
        else:
            continue

        if key.endswith("lat"):
            lonlat.read_direct(out_lats)
        elif key.endswith("lon"):
            lonlat.read_direct(out_lons)
        elif key.endswith("lat_reduced"):
            lat_reduced = lonlat[:]
            reduced_grid = True
        elif key.endswith("lon_reduced"):
            lon_reduced = lonlat[:]

    if reduced_grid:
        from geotiepoints import SatelliteInterpolator

        cols_full = np.arange(shape[1])
        rows_full = np.arange(shape[0])

        satint = SatelliteInterpolator((lon_reduced, lat_reduced),
                                       (row_indices,
                                        col_indices),
                                       (rows_full, cols_full))
        out_lons[:], out_lats[:] = satint.interpolate()

    new_mask = False
    # FIXME: this is to mask out the npp bowtie deleted pixels...
    if "NPP" in h5f.attrs['platform']:

        if shape[1] == 3200:  # M-bands:
            new_mask = np.zeros((16, 3200), dtype=bool)
            new_mask[0, :1008] = True
            new_mask[1, :640] = True
            new_mask[14, :640] = True
            new_mask[15, :1008] = True
            new_mask[14, 2560:] = True
            new_mask[1, 2560:] = True
            new_mask[0, 2192:] = True
            new_mask[15, 2192:] = True
            new_mask = np.tile(new_mask, (out_lons.shape[0] / 16, 1))
        elif shape[1] == 6400:  # I-bands:
            LOG.info(
                "PPS on I-band resolution. Mask out bow-tie deletion pixels")
            LOG.warning("Not yet supported...")
            new_mask = np.zeros((32, 6400), dtype=bool)
            new_mask[0:2, :2016] = True
            new_mask[0:2, 4384:] = True
            new_mask[2:4, :1280] = True
            new_mask[2:4, 5120:] = True
            new_mask[28:30, :1280] = True
            new_mask[28:30, 5120:] = True
            new_mask[30:32, :2016] = True
            new_mask[30:32, 4384:] = True
            new_mask = np.tile(new_mask, (out_lons.shape[0] / 32, 1))
        else:
            LOG.error("VIIRS shape not supported. " +
                      "No handling of bow-tie deletion pixels: shape = ", str(shape))

    out_mask[:] = np.logical_or(
        new_mask, np.logical_and(out_lats == fillvalue, out_lons == fillvalue))
    # new_mask, np.logical_and(out_lats <= fillvalue, out_lons <= fillvalue))

    h5f.close()
    if unzipped:
        os.remove(unzipped)


def globify_date(filename):
    """Replace date formats with single character wildcards"""
    filename = filename.replace("%Y", "????")
    filename = filename.replace("%m", "??")
    filename = filename.replace("%d", "??")
    filename = filename.replace("%H", "??")
    filename = filename.replace("%M", "??")
    filename = filename.replace("%S", "??")
    return filename



from mpop.imageo.geo_image import GeoImage

def s2_truecolor(self):
	
	self.check_channels('B02','B03','B04')
    
	ch1 = self['B04'].data
	ch2 = self['B03'].data
	ch3 = self['B02'].data

	img = GeoImage((ch1, ch2, ch3),
				self.area,
                                 self.time_slot,
                                 fill_value=None,
                                 mode="RGB")

	img.enhance(stretch="linear")
	#img.enhance(stretch="histogram")
	img.enhance(gamma=2.0)
    
	return img

s2_truecolor.prerequisites = set(['B02', 'B03','B04'])
msi=[s2_truecolor]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016.

# Author(s):

#   Christian Kliche <christian.kliche@ebp.de>

# This file is part of the mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.
'''
Module for writer option constants
'''
WR_OPT_NBITS = 'nbits'
WR_OPT_COMPRESSION = 'compression'
WR_OPT_BLOCKSIZE = 'blocksize'
WR_OPT_FILL_VALUE_SUBST = 'fill_value_subst'
